@@ -23,21 +23,6 @@
  linkages has a consistent post-processing behavior.  () compute the
  cost of the linkage. */
 
-
-static List_o_links *word_links[MAX_SENTENCE]; /* ptr to l.o.l. out of word */
-static int structure_violation;
-static int dfs_root_word[MAX_SENTENCE]; /* for the depth-first search */
-static int dfs_height[MAX_SENTENCE];    /* to determine the order to do the root word dfs */
-static int height_perm[MAX_SENTENCE];   /* permute the vertices from highest to lowest */
-
-/* The following three functions are all for computing the cost of and lists */
-static int visited[MAX_SENTENCE];
-static int and_element_sizes[MAX_SENTENCE];
-static int and_element[MAX_SENTENCE];
-static int N_and_elements;
-static int outside_word[MAX_SENTENCE];
-static int N_outside_words;
-
 typedef struct patch_element_struct Patch_element;
 struct patch_element_struct
 {
@@ -47,14 +32,30 @@ struct patch_element_struct
 	int newr;    /* the new value of the right end         */
 };
 
-static Patch_element patch_array[MAX_LINKS];
+struct analyze_context_s
+{
+	List_o_links *word_links[MAX_SENTENCE]; /* ptr to l.o.l. out of word */
+	int structure_violation;
+
+	int dfs_root_word[MAX_SENTENCE]; /* for the depth-first search */
+	int dfs_height[MAX_SENTENCE];    /* to determine the order to do the root word dfs */
+	int height_perm[MAX_SENTENCE];   /* permute the vertices from highest to lowest */
+
+	/* The following are all for computing the cost of and lists */
+	int visited[MAX_SENTENCE];
+	int and_element_sizes[MAX_SENTENCE];
+	int and_element[MAX_SENTENCE];
+	int N_and_elements;
+	int outside_word[MAX_SENTENCE];
+	int N_outside_words;
+	Patch_element patch_array[MAX_LINKS];
+};
 
 typedef struct DIS_node_struct DIS_node;
 typedef struct CON_node_struct CON_node;
 typedef struct CON_list_struct CON_list;
 typedef struct DIS_list_struct DIS_list;
 typedef struct Links_to_patch_struct Links_to_patch;
-typedef void (*void_returning_function)(void);
 
 struct DIS_node_struct {
 	CON_list * cl;     /* the list of children */
@@ -98,7 +99,7 @@ void zero_sublinkage(Sublinkage *s)
 static Sublinkage * x_create_sublinkage(Parse_info pi)
 {
 	Sublinkage *s = (Sublinkage *) xalloc (sizeof(Sublinkage));
-	s->link = (Link *) xalloc(MAX_LINKS*sizeof(Link));
+	s->link = (Link **) xalloc(MAX_LINKS*sizeof(Link *));
 	s->num_links = MAX_LINKS;
 
 	zero_sublinkage(s);
@@ -111,7 +112,7 @@ static Sublinkage * x_create_sublinkage(Parse_info pi)
 static Sublinkage * ex_create_sublinkage(Parse_info pi)
 {
 	Sublinkage *s = (Sublinkage *) exalloc (sizeof(Sublinkage));
-	s->link = (Link *) exalloc(pi->N_links*sizeof(Link));
+	s->link = (Link **) exalloc(pi->N_links*sizeof(Link *));
 	s->num_links = pi->N_links;
 
 	zero_sublinkage(s);
@@ -130,7 +131,7 @@ static void free_sublinkage(Sublinkage *s)
 	xfree(s, sizeof(Sublinkage));
 }
 
-static void replace_link_name(Link l, const char *s)
+static void replace_link_name(Link *l, const char *s)
 {
 	char * t;
 	exfree((char *) l->name, sizeof(char)*(strlen(l->name)+1));
@@ -139,7 +140,7 @@ static void replace_link_name(Link l, const char *s)
 	l->name = t;
 }
 
-static void copy_full_link(Link *dest, Link src)
+static void copy_full_link(Link **dest, Link *src)
 {
 	if (*dest != NULL) exfree_link(*dest);
 	*dest = excopy_link(src);
@@ -148,31 +149,36 @@ static void copy_full_link(Link *dest, Link src)
 /* end new code 9/97 ALB */
 
 
-/* Constructs a graph in the wordlinks array based on the contents of    */
-/* the global link_array.  Makes the wordlinks array point to a list of  */
-/* words neighboring each word (actually a list of links).  This is a     */
-/* directed graph, constructed for dealing with "and".  For a link in     */
-/* which the priorities are UP or DOWN_priority, the edge goes from the   */
-/* one labeled DOWN to the one labeled UP.                                */
-/* Don't generate links edges for the bogus comma connectors              */
-static void build_digraph(Parse_info pi, List_o_links **wordlinks)
+/** 
+ * Constructs a graph in the wordlinks array based on the contents of
+ * the global link_array.  Makes the wordlinks array point to a list of
+ * words neighboring each word (actually a list of links).  This is a
+ * directed graph, constructed for dealing with "and".  For a link in
+ * which the priorities are UP or DOWN_priority, the edge goes from the
+ * one labeled DOWN to the one labeled UP.
+ * Don't generate links edges for the bogus comma connectors.
+ */
+static void build_digraph(analyze_context_t *actx, Parse_info pi)
 {
 	int i, link, N_fat;
-	Link lp;
+	Link *lp;
 	List_o_links * lol;
 	N_fat = 0;
-	for (i=0; i<pi->N_words; i++) {
-		wordlinks[i] = NULL;
+
+	for (i=0; i<pi->N_words; i++)
+	{
+		actx->word_links[i] = NULL;
 	}
-	for (link=0; link<pi->N_links; link++) {
+	for (link=0; link<pi->N_links; link++)
+	{
 		lp = &(pi->link_array[link]);
 		i = lp->lc->label;
 		if (i < NORMAL_LABEL) {   /* one of those special links for either-or, etc */
 			continue;
 		}
 		lol = (List_o_links *) xalloc(sizeof(List_o_links));
-		lol->next = wordlinks[lp->l];
-		wordlinks[lp->l] = lol;
+		lol->next = actx->word_links[lp->l];
+		actx->word_links[lp->l] = lol;
 		lol->link = link;
 		lol->word = lp->r;
 		i = lp->lc->priority;
@@ -184,8 +190,8 @@ static void build_digraph(Parse_info pi, List_o_links **wordlinks)
 			lol->dir = -1;
 		}
 		lol = (List_o_links *) xalloc(sizeof(List_o_links));
-		lol->next = wordlinks[lp->r];
-		wordlinks[lp->r] = lol;
+		lol->next = actx->word_links[lp->r];
+		actx->word_links[lp->r] = lol;
 		lol->link = link;
 		lol->word = lp->l;
 		i = lp->rc->priority;
@@ -213,23 +219,25 @@ static int is_CON_word(int w, List_o_links **wordlinks)
 	return FALSE;
 }
 
-static DIS_node * build_DIS_node(int, List_o_links **);
+static DIS_node * build_DIS_node(analyze_context_t*, int);
 
 /** 
  * This word is a CON word (has fat links down).  Build the tree for it.
  */
-static CON_node * build_CON_node(int w, List_o_links **wordlinks) 
+static CON_node * build_CON_node(analyze_context_t *actx, int w)
 {
 	List_o_links * lol;
 	CON_node * a;
 	DIS_list * d, *dx;
 	d = NULL;
-	for (lol = wordlinks[w]; lol!=NULL; lol = lol->next) {
-		if (lol->dir == 1) {
+	for (lol = actx->word_links[w]; lol!=NULL; lol = lol->next)
+	{
+		if (lol->dir == 1)
+		{
 			dx = (DIS_list *) xalloc (sizeof (DIS_list));
 			dx->next = d;
 			d = dx;
-			d->dn = build_DIS_node(lol->word, wordlinks);
+			d->dn = build_DIS_node(actx, lol->word);
 		}
 	}
 	a = (CON_node *) xalloc(sizeof (CON_node));
@@ -248,36 +256,46 @@ static CON_node * build_CON_node(int w, List_o_links **wordlinks)
  * linkages that have improper structure.  Fortunately, they
  * seem to be rather rare.
  */
-static CON_list * c_dfs(int w, DIS_node * start_dn, CON_list * c,
-                        List_o_links **wordlinks)
+static CON_list * c_dfs(analyze_context_t *actx,
+                        int w, DIS_node * start_dn, CON_list * c)
 {
 	CON_list *cx;
 	List_o_links * lol, *lolx;
-	if (dfs_root_word[w] != -1) {
-		if (dfs_root_word[w] != start_dn->word) {
-			structure_violation = TRUE;
+	if (actx->dfs_root_word[w] != -1)
+	{
+		if (actx->dfs_root_word[w] != start_dn->word)
+		{
+			actx->structure_violation = TRUE;
 		}
 		return c;
 	}
-	dfs_root_word[w] = start_dn->word;
-	for (lol=wordlinks[w]; lol != NULL; lol = lol->next) {
-		if (lol->dir < 0) {  /* a backwards link */
-			if (dfs_root_word[lol->word] == -1) {
-				structure_violation = TRUE;
+	actx->dfs_root_word[w] = start_dn->word;
+	for (lol = actx->word_links[w]; lol != NULL; lol = lol->next)
+	{
+		if (lol->dir < 0)  /* a backwards link */
+		{
+			if (actx->dfs_root_word[lol->word] == -1)
+			{
+				actx->structure_violation = TRUE;
 			}
-		} else if (lol->dir == 0) {
+		}
+		else if (lol->dir == 0)
+		{
 			lolx = (List_o_links *) xalloc(sizeof(List_o_links));
 			lolx->next = start_dn->lol;
 			lolx->link = lol->link;
 			start_dn->lol = lolx;
-			c = c_dfs(lol->word, start_dn, c, wordlinks);
+			c = c_dfs(actx, lol->word, start_dn, c);
 		}
 	}
-	if (is_CON_word(w, wordlinks)) {  /* if the current node is CON, put it first */
+
+	/* if the current node is CON, put it first */
+	if (is_CON_word(w, actx->word_links))
+	{
 		cx = (CON_list *) xalloc(sizeof(CON_list));
 		cx->next = c;
 		c = cx;
-		c->cn = build_CON_node(w, wordlinks);
+		c->cn = build_CON_node(actx, w);
 	}
 	return c;
 }
@@ -287,40 +305,49 @@ static CON_list * c_dfs(int w, DIS_node * start_dn, CON_list * c,
  * region reachable via thin links, and put all reachable nodes with fat
  * links out of them in its list of children.
  */
-static DIS_node * build_DIS_node(int w, List_o_links **wordlinks)
+static DIS_node * build_DIS_node(analyze_context_t *actx, 
+                                 int w)
 {
 	DIS_node * dn;
 	dn = (DIS_node *) xalloc(sizeof (DIS_node));
 	dn->word = w;   /* must do this before dfs so it knows the start word */
 	dn->lol = NULL;
-	dn->cl = c_dfs(w, dn, NULL, wordlinks);
+	dn->cl = c_dfs(actx, w, dn, NULL);
 	return dn;
 }
 
-static void height_dfs(int w, int height, List_o_links **wordlinks)
+static void height_dfs(analyze_context_t *actx, int w, int height)
 {
 	List_o_links * lol;
-	if (dfs_height[w] != 0) return;
-	dfs_height[w] = height;
-	for (lol=wordlinks[w]; lol != NULL; lol = lol->next) {
+	if (actx->dfs_height[w] != 0) return;
+
+	actx->dfs_height[w] = height;
+
+	for (lol = actx->word_links[w]; lol != NULL; lol = lol->next)
+	{
 		/* The dir is 1 for a down link. */
-		height_dfs(lol->word, height - lol->dir, wordlinks);
+		height_dfs(actx, lol->word, height - lol->dir);
 	}
 }
 
-static int comp_height(int *a, int *b) {
-	return dfs_height[*b] - dfs_height[*a];
-}
-
-#define COMPARE_TYPE int (*)(const void *, const void *)
-
-static DIS_node * build_DIS_CON_tree(Parse_info pi, List_o_links **wordlinks)
+static DIS_node * build_DIS_CON_tree(analyze_context_t *actx, Parse_info pi)
 {
 	int xw, w;
 	DIS_node * dnroot, * dn;
 	CON_list * child, * xchild;
 	List_o_links * lol, * xlol;
 
+	/* Declare the compare function here, since it needs access to the
+	 * stack varaiable actx, which would otherwise need to be delcared
+	 * global (and thus not thread-safe).
+	 */
+	int comp_height(const void *va, const void *vb)
+	{
+		const int *a = va; 
+		const int *b = vb;
+		return actx->dfs_height[*b] - actx->dfs_height[*a];
+	};
+
 	/* The algorithm used here to build the DIS_CON tree depends on
 	   the search percolating down from the "top" of the tree.  The
 	   original version of this started its search at the wall.  This
@@ -332,23 +359,31 @@ static DIS_node * build_DIS_CON_tree(Parse_info pi, List_o_links **wordlinks)
 	   dfs_height[] for this.
 	 */
 
-	for (w=0; w < pi->N_words; w++) dfs_height[w] = 0;
-	for (w=0; w < pi->N_words; w++) height_dfs(w, MAX_SENTENCE, wordlinks);
+	for (w=0; w < pi->N_words; w++) actx->dfs_height[w] = 0;
+	for (w=0; w < pi->N_words; w++) height_dfs(actx, w, MAX_SENTENCE);
 
-	for (w=0; w < pi->N_words; w++) height_perm[w] = w;
-	qsort(height_perm, pi->N_words, sizeof(height_perm[0]), (COMPARE_TYPE) comp_height);
+	for (w=0; w < pi->N_words; w++) actx->height_perm[w] = w;
 
-	for (w=0; w<pi->N_words; w++) dfs_root_word[w] = -1;
+	qsort(actx->height_perm, pi->N_words, 
+		sizeof(actx->height_perm[0]), comp_height);
+
+	for (w=0; w<pi->N_words; w++) actx->dfs_root_word[w] = -1;
 
 	dnroot = NULL;
-	for (xw=0; xw < pi->N_words; xw++) {
-		w = height_perm[xw];
-		if (dfs_root_word[w] == -1) {
-			dn = build_DIS_node(w, wordlinks);
-			if (dnroot == NULL) {
+	for (xw=0; xw < pi->N_words; xw++)
+	{
+		w = actx->height_perm[xw];
+		if (actx->dfs_root_word[w] == -1)
+		{
+			dn = build_DIS_node(actx, w);
+			if (dnroot == NULL)
+			{
 				dnroot = dn;
-			} else {
-				for (child = dn->cl; child != NULL; child = xchild) {
+			}
+			else
+			{
+				for (child = dn->cl; child != NULL; child = xchild)
+				{
 					xchild = child->next;
 					child->next = dnroot->cl;
 					dnroot->cl = child;
@@ -366,11 +401,14 @@ static DIS_node * build_DIS_CON_tree(Parse_info pi, List_o_links **wordlinks)
 }
 
 static int advance_CON(CON_node *);
-static int advance_DIS(DIS_node * dn) {
-/* Cycically advance the current state of this DIS node.
-   If it's now at the beginning of its cycle" return FALSE;
-   Otherwise return TRUE;
-*/
+
+/**
+ * Cycically advance the current state of this DIS node.
+ * If it's now at the beginning of its cycle" return FALSE;
+ * Otherwise return TRUE;
+ */
+static int advance_DIS(DIS_node * dn)
+{
 	CON_list * cl;
 	for (cl = dn->cl; cl!=NULL; cl=cl->next) {
 		if (advance_CON(cl->cn)) {
@@ -380,11 +418,13 @@ static int advance_DIS(DIS_node * dn) {
 	return FALSE;
 }
 
-static int advance_CON(CON_node * cn) {
-/* Cycically advance the current state of this CON node.
-   If it's now at the beginning of its cycle return FALSE;
-   Otherwise return TRUE;
-*/
+/**
+ * Cycically advance the current state of this CON node.
+ * If it's now at the beginning of its cycle return FALSE;
+ * Otherwise return TRUE;
+ */
+static int advance_CON(CON_node * cn)
+{
 	if (advance_DIS(cn->current->dn)) {
 		return TRUE;
 	} else {
@@ -398,31 +438,37 @@ static int advance_CON(CON_node * cn) {
 	}
 }
 
-static void fill_patch_array_CON(CON_node *, Links_to_patch *, List_o_links**);
+static void fill_patch_array_CON(analyze_context_t *, CON_node *, Links_to_patch *);
 
 /**
  * Patches up appropriate links in the patch_array for this DIS_node
  * and this patch list.
  */
-static void fill_patch_array_DIS(DIS_node * dn, Links_to_patch * ltp,
-                                 List_o_links **wordlinks)
+static void fill_patch_array_DIS(analyze_context_t *actx, 
+                                 DIS_node * dn, Links_to_patch * ltp)
 {
 	CON_list * cl;
 	List_o_links * lol;
 	Links_to_patch * ltpx;
 
-	for (lol=dn->lol; lol != NULL; lol=lol->next) {
-		patch_array[lol->link].used = TRUE;
+	for (lol = dn->lol; lol != NULL; lol=lol->next)
+	{
+		actx->patch_array[lol->link].used = TRUE;
 	}
 
-	if ((dn->cl == NULL) || (dn->cl->cn->word != dn->word)) {
-		for (; ltp != NULL; ltp = ltpx) {
+	if ((dn->cl == NULL) || (dn->cl->cn->word != dn->word))
+	{
+		for (; ltp != NULL; ltp = ltpx)
+		{
 			ltpx = ltp->next;
-			patch_array[ltp->link].changed = TRUE;
-			if (ltp->dir == 'l') {
-				patch_array[ltp->link].newl = dn->word;
-			} else {
-				patch_array[ltp->link].newr = dn->word;
+			actx->patch_array[ltp->link].changed = TRUE;
+			if (ltp->dir == 'l')
+			{
+				actx->patch_array[ltp->link].newl = dn->word;
+			}
+			else
+			{
+				actx->patch_array[ltp->link].newr = dn->word;
 			}
 			xfree((void *) ltp, sizeof(Links_to_patch));
 		}
@@ -430,20 +476,23 @@ static void fill_patch_array_DIS(DIS_node * dn, Links_to_patch * ltp,
 
 	/* ltp != NULL at this point means that dn has child which is a cn
 	   which is the same word */
-	for (cl=dn->cl; cl!=NULL; cl=cl->next) {
-		fill_patch_array_CON(cl->cn, ltp, wordlinks);
+	for (cl = dn->cl; cl != NULL; cl = cl->next)
+	{
+		fill_patch_array_CON(actx, cl->cn, ltp);
 		ltp = NULL;
 	}
 }
 
-static void fill_patch_array_CON(CON_node * cn, Links_to_patch * ltp,
-                                 List_o_links **wordlinks)
+static void fill_patch_array_CON(analyze_context_t *actx, 
+                                 CON_node * cn, Links_to_patch * ltp)
 {
 	List_o_links * lol;
 	Links_to_patch *ltpx;
 
-	for (lol=wordlinks[cn->word]; lol != NULL; lol = lol->next) {
-		if (lol->dir == 0) {
+	for (lol = actx->word_links[cn->word]; lol != NULL; lol = lol->next)
+	{
+		if (lol->dir == 0)
+		{
 			ltpx = (Links_to_patch *) xalloc(sizeof(Links_to_patch));
 			ltpx->next = ltp;
 			ltp = ltpx;
@@ -455,25 +504,27 @@ static void fill_patch_array_CON(CON_node * cn, Links_to_patch * ltp,
 			}
 		}
 	}
-	fill_patch_array_DIS(cn->current->dn, ltp, wordlinks);
+	fill_patch_array_DIS(actx, cn->current->dn, ltp);
 }
 
-static void free_digraph(Parse_info pi, List_o_links **wordlinks)
+static void free_digraph(analyze_context_t *actx, Parse_info pi)
 {
-  List_o_links * lol, *lolx;
-  int i;
-  for (i=0; i<pi->N_words; i++)
+	List_o_links * lol, *lolx;
+	int i;
+	for (i = 0; i < pi->N_words; i++)
 	{
-	  for (lol=wordlinks[i]; lol!=NULL; lol=lolx)
+		for (lol = actx->word_links[i]; lol != NULL; lol = lolx)
 		{
-		  lolx = lol->next;
-		  xfree((void *) lol, sizeof(List_o_links));
+			lolx = lol->next;
+			xfree((void *) lol, sizeof(List_o_links));
 		}
 	}
 }
 
 static void free_CON_tree(CON_node *);
-static void free_DIS_tree(DIS_node * dn) {
+
+static void free_DIS_tree(DIS_node * dn)
+{
 	List_o_links * lol, *lolx;
 	CON_list *cl, *clx;
 	for (lol=dn->lol; lol!=NULL; lol=lolx) {
@@ -488,7 +539,8 @@ static void free_DIS_tree(DIS_node * dn) {
 	xfree((void *) dn, sizeof(DIS_node));
 }
 
-static void free_CON_tree(CON_node * cn) {
+static void free_CON_tree(CON_node * cn)
+{
 	DIS_list *dl, *dlx;
 	for (dl = cn->dl; dl!=NULL; dl=dlx) {
 		dlx = dl->next;
@@ -499,42 +551,51 @@ static void free_CON_tree(CON_node * cn) {
 }
 
 /** scope out this and element */
-static void and_dfs_full(int w, List_o_links **wordlinks)
+static void and_dfs_full(analyze_context_t *actx, int w)
 {
 	List_o_links *lol;
-	if (visited[w]) return;
-	visited[w] = TRUE;
-	and_element_sizes[N_and_elements]++;
+	if (actx->visited[w]) return;
+	actx->visited[w] = TRUE;
+	actx->and_element_sizes[actx->N_and_elements]++;
 
-	for (lol = wordlinks[w]; lol != NULL; lol = lol->next) {
-		if (lol->dir >= 0) {
-			and_dfs_full(lol->word, wordlinks);
+	for (lol = actx->word_links[w]; lol != NULL; lol = lol->next)
+	{
+		if (lol->dir >= 0)
+		{
+			and_dfs_full(actx, lol->word);
 		}
 	}
 }
 
 /** get down the tree past all the commas */
-static void and_dfs_commas(Sentence sent, int w, List_o_links **wordlinks)
+static void and_dfs_commas(analyze_context_t *actx, Sentence sent, int w)
 {
 	List_o_links *lol;
-	if (visited[w]) return;
-	visited[w] = TRUE;
-	for (lol = wordlinks[w]; lol != NULL; lol = lol->next) {
-		if (lol->dir == 1) {
-			 /* we only consider UP or DOWN priority links here */
+	if (actx->visited[w]) return;
 
-			if (strcmp(sent->word[lol->word].string, ",")==0) {
-					/* pointing to a comma */
-				and_dfs_commas(sent, lol->word, wordlinks);
-			} else {
-				and_element[N_and_elements]=lol->word;
-				and_dfs_full(lol->word, wordlinks);
-				N_and_elements++;
+	actx->visited[w] = TRUE;
+
+	for (lol = actx->word_links[w]; lol != NULL; lol = lol->next)
+	{
+		/* we only consider UP or DOWN priority links here */
+		if (lol->dir == 1)
+		{
+			if (strcmp(sent->word[lol->word].string, ",") == 0)
+			{
+				/* pointing to a comma */
+				and_dfs_commas(actx, sent, lol->word);
+			}
+			else
+			{
+				actx->and_element[actx->N_and_elements] = lol->word;
+				and_dfs_full(actx, lol->word);
+				actx->N_and_elements++;
 			}
 		}
-		if (lol->dir == 0) {
-		  outside_word[N_outside_words]=lol->word;
-		  N_outside_words++;
+		if (lol->dir == 0)
+		{
+		  actx->outside_word[actx->N_outside_words] = lol->word;
+		  actx->N_outside_words++;
 		}
 	}
 }
@@ -544,7 +605,7 @@ static void and_dfs_commas(Sentence sent, int w, List_o_links **wordlinks)
  * in the length of and-list elements. It also computes other
  * information used to construct the "andlist" structure of linkage_info.
  */
-static Andlist * build_andlist(Sentence sent, List_o_links **wordlinks)
+static Andlist * build_andlist(analyze_context_t *actx, Sentence sent)
 {
 	int w, i, min, max, j, cost;
 	char * s;
@@ -554,39 +615,49 @@ static Andlist * build_andlist(Sentence sent, List_o_links **wordlinks)
 	old_andlist = NULL;
 	cost = 0;
 
-	for(w = 0; w<pi->N_words; w++) {
+	for(w = 0; w<pi->N_words; w++)
+	{
 		s = sent->word[w].string;
-		if (sent->is_conjunction[w]) {
-			N_and_elements = 0;
-			N_outside_words = 0;
-			for(i=0; i<pi->N_words; i++) {
-				visited[i] = FALSE;
-				and_element_sizes[i] = 0;
+		if (sent->is_conjunction[w])
+		{
+			actx->N_and_elements = 0;
+			actx->N_outside_words = 0;
+			for(i=0; i<pi->N_words; i++)
+			{
+				actx->visited[i] = FALSE;
+				actx->and_element_sizes[i] = 0;
 			}
 			if (sent->dict->left_wall_defined)
-				visited[0] = TRUE;
-			and_dfs_commas(sent, w, wordlinks);
-			if(N_and_elements == 0) continue;
-			  new_andlist = (Andlist *) xalloc(sizeof(Andlist));
-			new_andlist->num_elements = N_and_elements;
-			new_andlist->num_outside_words = N_outside_words;
-			for (i=0; i<N_and_elements; i++) {
-			  new_andlist->element[i] = and_element[i];
+				actx->visited[0] = TRUE;
+
+			and_dfs_commas(actx, sent, w);
+			if (actx->N_and_elements == 0) continue;
+
+			new_andlist = (Andlist *) xalloc(sizeof(Andlist));
+			new_andlist->num_elements = actx->N_and_elements;
+			new_andlist->num_outside_words = actx->N_outside_words;
+
+			for (i=0; i < actx->N_and_elements; i++)
+			{
+				new_andlist->element[i] = actx->and_element[i];
 			}
-			for (i=0; i<N_outside_words; i++) {
-			  new_andlist->outside_word[i] = outside_word[i];
+			for (i=0; i < actx->N_outside_words; i++)
+			{
+				new_andlist->outside_word[i] = actx->outside_word[i];
 			}
 			new_andlist->conjunction = w;
 			new_andlist->next = old_andlist;
 			old_andlist = new_andlist;
 
-			if (N_and_elements > 0) {
-				min=MAX_SENTENCE;
-				max=0;
-				for (i=0; i<N_and_elements; i++) {
-					j = and_element_sizes[i];
-					if (j < min) min=j;
-					if (j > max) max=j;
+			if (actx->N_and_elements > 0)
+			{
+				min = MAX_SENTENCE;
+				max = 0;
+				for (i=0; i < actx->N_and_elements; i++)
+				{
+					j = actx->and_element_sizes[i];
+					if (j < min) min = j;
+					if (j > max) max = j;
 				}
 				cost += max-min;
 			}
@@ -596,26 +667,20 @@ static Andlist * build_andlist(Sentence sent, List_o_links **wordlinks)
 	return old_andlist;
 }
 
-#ifdef DEAD_CODE
-static void islands_dfs(int w, List_o_links **wordlinks)
+/** 
+ * This function defines the cost of a link as a function of its length.
+ */
+static inline int cost_for_length(int length)
 {
-	List_o_links *lol;
-	if (visited[w]) return;
-	visited[w] = TRUE;
-	for (lol = wordlinks[w]; lol != NULL; lol = lol->next) {
-	  islands_dfs(lol->word, wordlinks);
-	}
-}
-#endif /* DEAD_CODE */
-
-static int cost_for_length(int length) {
-/* this function defines the cost of a link as a function of its length */
 	  return length-1;
 }
 
-static int link_cost(Parse_info pi) {
-/* computes the cost of the current parse of the current sentence */
-/* due to the length of the links                                 */
+/**
+ * Computes the cost of the current parse of the current sentence,
+ * due to the length of the links.
+ */
+static int link_cost(Parse_info pi)
+{
 	int lcost, i;
 	lcost =  0;
 	for (i=0; i<pi->N_links; i++) {
@@ -624,13 +689,15 @@ static int link_cost(Parse_info pi) {
 	return lcost;
 }
 
-static int null_cost(Parse_info pi) {
+static int null_cost(Parse_info pi)
+{
   /* computes the number of null links in the linkage */
   /* No one seems to care about this -- ALB */
   return 0;
 }
 
-static int unused_word_cost(Parse_info pi) {
+static int unused_word_cost(Parse_info pi)
+{
 	int lcost, i;
 	lcost =  0;
 	for (i=0; i<pi->N_words; i++)
@@ -639,7 +706,8 @@ static int unused_word_cost(Parse_info pi) {
 }
 
 
-static int disjunct_cost(Parse_info pi) {
+static int disjunct_cost(Parse_info pi)
+{
 /* computes the cost of the current parse of the current sentence     */
 /* due to the cost of the chosen disjuncts                            */
 	int lcost, i;
@@ -718,7 +786,7 @@ static void compute_pp_link_names(Sentence sent, Sublinkage *sublinkage)
 	for (i=0; i<pi->N_links; i++)
 	  {
 		if (sublinkage->link[i]->l == -1) continue;
-		if (!x_match(sublinkage->link[i]->lc, sublinkage->link[i]->rc))
+		if (!x_match(sent, sublinkage->link[i]->lc, sublinkage->link[i]->rc))
 		  replace_link_name(sublinkage->link[i], pi->link_array[i].name);
 		else
 		  {
@@ -733,6 +801,22 @@ static void compute_pp_link_names(Sentence sent, Sublinkage *sublinkage)
 
 /********************** exported functions *****************************/
 
+void init_analyze(Sentence s)
+{
+	analyze_context_t *actx;
+
+	actx = (analyze_context_t *) malloc (sizeof(analyze_context_t));
+	actx->structure_violation = FALSE;
+
+	s->analyze_ctxt = actx;
+}
+
+void free_analyze(Sentence s)
+{
+	if (s->analyze_ctxt != NULL) free(s->analyze_ctxt);
+	s->analyze_ctxt = NULL;
+}
+
 /**
  * This uses link_array.  It enumerates and post-processes
  * all the linkages represented by this one.  We know this contains
@@ -750,14 +834,16 @@ Linkage_info analyze_fat_linkage(Sentence sent, Parse_Options opts, int analyze_
 	PP_node accum;			   /* for domain ancestry check */
 	D_type_list * dtl0, * dtl1;  /* for domain ancestry check */
 
+	analyze_context_t *actx = sent->analyze_ctxt;
+
 	sublinkage = x_create_sublinkage(pi);
 	postprocessor = sent->dict->postprocessor;
-	build_digraph(pi, word_links);
-	structure_violation = FALSE;
-	d_root = build_DIS_CON_tree(pi, word_links); /* may set structure_violation to TRUE */
+	build_digraph(actx, pi);
+	actx->structure_violation = FALSE;
+	d_root = build_DIS_CON_tree(actx, pi); /* may set structure_violation to TRUE */
 
 	li.N_violations = 0;
-	li.improper_fat_linkage = structure_violation;
+	li.improper_fat_linkage = actx->structure_violation;
 	li.inconsistent_domains = FALSE;
 	li.unused_word_cost = unused_word_cost(sent->parse_info);
 	li.disjunct_cost = disjunct_cost(pi);
@@ -766,16 +852,18 @@ Linkage_info analyze_fat_linkage(Sentence sent, Parse_Options opts, int analyze_
 	li.and_cost = 0;
 	li.andlist = NULL;
 
-	if (structure_violation) {
+	if (actx->structure_violation)
+	{
 		li.N_violations++;
 		free_sublinkage(sublinkage);
-		free_digraph(pi, word_links);
+		free_digraph(actx, pi);
 		free_DIS_tree(d_root);
 		return li;
 	}
 
-	if (analyze_pass==PP_SECOND_PASS) {
-	  li.andlist = build_andlist(sent, word_links);
+	if (analyze_pass==PP_SECOND_PASS)
+	{
+	  li.andlist = build_andlist(actx, sent);
 	  li.and_cost = li.andlist->cost;
 	}
 	else li.and_cost = 0;
@@ -784,22 +872,28 @@ Linkage_info analyze_fat_linkage(Sentence sent, Parse_Options opts, int analyze_
 
 	for (i=0; i<pi->N_links; i++) accum.d_type_array[i] = NULL;
 
-	for (;;) {		/* loop through all the sub linkages */
-		for (i=0; i<pi->N_links; i++) {
-			patch_array[i].used = patch_array[i].changed = FALSE;
-			patch_array[i].newl = pi->link_array[i].l;
-			patch_array[i].newr = pi->link_array[i].r;
+	/* loop through all the sub linkages */
+	for (;;)
+	{
+		for (i=0; i<pi->N_links; i++)
+		{
+			actx->patch_array[i].used = actx->patch_array[i].changed = FALSE;
+			actx->patch_array[i].newl = pi->link_array[i].l;
+			actx->patch_array[i].newr = pi->link_array[i].r;
 			copy_full_link(&sublinkage->link[i], &(pi->link_array[i]));
 		}
-		fill_patch_array_DIS(d_root, NULL, word_links);
+		fill_patch_array_DIS(actx, d_root, NULL);
 
-		for (i=0; i<pi->N_links; i++) {
-			if (patch_array[i].changed || patch_array[i].used) {
-				sublinkage->link[i]->l = patch_array[i].newl;
-				sublinkage->link[i]->r = patch_array[i].newr;
+		for (i=0; i<pi->N_links; i++)
+		{
+			if (actx->patch_array[i].changed || actx->patch_array[i].used)
+			{
+				sublinkage->link[i]->l = actx->patch_array[i].newl;
+				sublinkage->link[i]->r = actx->patch_array[i].newr;
 			}
-			else if ((dfs_root_word[pi->link_array[i].l] != -1) &&
-					 (dfs_root_word[pi->link_array[i].r] != -1)) {
+			else if ((actx->dfs_root_word[pi->link_array[i].l] != -1) &&
+					 (actx->dfs_root_word[pi->link_array[i].r] != -1))
+			{
 				sublinkage->link[i]->l = -1;
 			}
 		}
@@ -857,7 +951,7 @@ Linkage_info analyze_fat_linkage(Sentence sent, Parse_Options opts, int analyze_
 	   (verbosity > 3) && should_print_messages)
 	   printf("P.P. violation in one part of conjunction.\n"); */
 	free_sublinkage(sublinkage);
-	free_digraph(pi, word_links);
+	free_digraph(actx, pi);
 	free_DIS_tree(d_root);
 	return li;
 }
@@ -875,29 +969,33 @@ Linkage_info analyze_thin_linkage(Sentence sent, Parse_Options opts, int analyze
 	Postprocessor * postprocessor;
 	Sublinkage *sublinkage;
 	Parse_info pi = sent->parse_info;
+	analyze_context_t *actx = sent->analyze_ctxt;
 
-	build_digraph(pi, word_links);
+	build_digraph(actx, pi);
 	memset(&li, 0, sizeof(li));
 
 	sublinkage = x_create_sublinkage(pi);
 	postprocessor = sent->dict->postprocessor;
 
 	compute_link_names(sent);
-	for (i=0; i<pi->N_links; i++) {
+	for (i=0; i<pi->N_links; i++)
+	{
 	  copy_full_link(&(sublinkage->link[i]), &(pi->link_array[i]));
 	}
 
-	if (analyze_pass==PP_FIRST_PASS) {
+	if (analyze_pass==PP_FIRST_PASS)
+	{
 		post_process_scan_linkage(postprocessor, opts, sent, sublinkage);
 		free_sublinkage(sublinkage);
-		free_digraph(pi, word_links);
+		free_digraph(actx, pi);
 		return li;
 	}
 
-	/* The code below can be used to generate the "islands" array. For this to work,
-	 * however, you have to call "build_digraph" first (as in analyze_fat_linkage).
-	 * and then "free_digraph". For some reason this causes a space leak. */
-
+	/* The code below can be used to generate the "islands" array.
+	 * For this to work, however, you have to call "build_digraph"
+	 * first (as in analyze_fat_linkage). and then "free_digraph".
+	 * For some reason this causes a space leak. XXX !! FIXME!?
+	 */
 	pp = post_process(postprocessor, opts, sent, sublinkage, TRUE);
 
 	li.N_violations = 0;
@@ -910,14 +1008,17 @@ Linkage_info analyze_thin_linkage(Sentence sent, Parse_Options opts, int analyze
 	li.link_cost = link_cost(pi);
 	li.andlist = NULL;
 
-	if (pp==NULL) {
+	if (pp == NULL)
+	{
 		if (postprocessor != NULL) li.N_violations = 1;
-	} else if (pp->violation!=NULL) {
+	}
+	else if (pp->violation != NULL)
+	{
 		li.N_violations++;
 	}
 
 	free_sublinkage(sublinkage);
-	free_digraph(pi, word_links);
+	free_digraph(actx, pi);
 	return li;
 }
 
@@ -957,14 +1058,18 @@ void extract_fat_linkage(Sentence sent, Parse_Options opts, Linkage linkage)
 	Sublinkage * sublinkage;
 	Parse_info pi = sent->parse_info;
 
-	sublinkage = x_create_sublinkage(pi);
-	build_digraph(pi, word_links);
-	structure_violation = FALSE;
-	d_root = build_DIS_CON_tree(pi, word_links);
+	analyze_context_t *actx = sent->analyze_ctxt;
 
-	if (structure_violation) {
+	sublinkage = x_create_sublinkage(pi);
+	build_digraph(actx, pi);
+	actx->structure_violation = FALSE;
+	d_root = build_DIS_CON_tree(actx, pi);
+
+	if (actx->structure_violation)
+	{
 		compute_link_names(sent);
-		for (i=0; i<pi->N_links; i++) {
+		for (i=0; i<pi->N_links; i++)
+		{
 			copy_full_link(&sublinkage->link[i],&(pi->link_array[i]));
 		}
 
@@ -972,12 +1077,13 @@ void extract_fat_linkage(Sentence sent, Parse_Options opts, Linkage linkage)
 		linkage->sublinkage = ex_create_sublinkage(pi);
 
 		/* This will have fat links! */
-		for (i=0; i<pi->N_links; ++i) {
+		for (i=0; i<pi->N_links; ++i)
+		{
 			linkage->sublinkage->link[i] = excopy_link(sublinkage->link[i]);
 		}
 
 		free_sublinkage(sublinkage);
-		free_digraph(pi, word_links);
+		free_digraph(actx, pi);
 		free_DIS_tree(d_root);
 		return;
 	}
@@ -1002,21 +1108,27 @@ void extract_fat_linkage(Sentence sent, Parse_Options opts, Linkage linkage)
 	compute_link_names(sent);
 
 	num_sublinkages = 0;
-	for (;;) {
-		for (i=0; i<pi->N_links; i++) {
-			patch_array[i].used = patch_array[i].changed = FALSE;
-			patch_array[i].newl = pi->link_array[i].l;
-			patch_array[i].newr = pi->link_array[i].r;
+	for (;;)
+	{
+		for (i=0; i<pi->N_links; i++)
+		{
+			actx->patch_array[i].used = actx->patch_array[i].changed = FALSE;
+			actx->patch_array[i].newl = pi->link_array[i].l;
+			actx->patch_array[i].newr = pi->link_array[i].r;
 			copy_full_link(&sublinkage->link[i], &(pi->link_array[i]));
 		}
-		fill_patch_array_DIS(d_root, NULL, word_links);
+		fill_patch_array_DIS(actx, d_root, NULL);
 
-		for (i=0; i<pi->N_links; i++) {
-			if (patch_array[i].changed || patch_array[i].used) {
-				sublinkage->link[i]->l = patch_array[i].newl;
-				sublinkage->link[i]->r = patch_array[i].newr;
-			} else if ((dfs_root_word[pi->link_array[i].l] != -1) &&
-					   (dfs_root_word[pi->link_array[i].r] != -1)) {
+		for (i=0; i<pi->N_links; i++)
+		{
+			if (actx->patch_array[i].changed || actx->patch_array[i].used)
+			{
+				sublinkage->link[i]->l = actx->patch_array[i].newl;
+				sublinkage->link[i]->r = actx->patch_array[i].newr;
+			}
+			else if ((actx->dfs_root_word[pi->link_array[i].l] != -1) &&
+					   (actx->dfs_root_word[pi->link_array[i].r] != -1))
+			{
 				sublinkage->link[i]->l = -1;
 			}
 		}
@@ -1033,7 +1145,7 @@ void extract_fat_linkage(Sentence sent, Parse_Options opts, Linkage linkage)
 
 		linkage->sublinkage[num_sublinkages].num_links = N_thin_links;
 		linkage->sublinkage[num_sublinkages].link =
-			(Link *) exalloc(sizeof(Link)*N_thin_links);
+			(Link **) exalloc(sizeof(Link *)*N_thin_links);
 		linkage->sublinkage[num_sublinkages].pp_info = NULL;
 		linkage->sublinkage[num_sublinkages].violation = NULL;
 
@@ -1049,7 +1161,7 @@ void extract_fat_linkage(Sentence sent, Parse_Options opts, Linkage linkage)
 	}
 
 	free_sublinkage(sublinkage);
-	free_digraph(pi, word_links);
+	free_digraph(actx, pi);
 	free_DIS_tree(d_root);
 }
 
