@@ -47,6 +47,7 @@ Parse_Options parse_options_create(void)
 {
 	Parse_Options po;
 
+	init_memusage();
 	po = (Parse_Options) xalloc(sizeof(struct Parse_Options_s));
 
 	/* Here's where the values are initialized */
@@ -62,7 +63,7 @@ Parse_Options parse_options_create(void)
 	po->short_length = 6;
 	po->all_short = FALSE;
 	po->twopass_length = 30;
-	po->max_sentence_length = 70;
+	po->max_sentence_length = 170;
 	po->resources = resources_create();
 	po->display_short = TRUE;
 	po->display_word_subscripts = TRUE;
@@ -83,7 +84,8 @@ Parse_Options parse_options_create(void)
 	return po;
 }
 
-int parse_options_delete(Parse_Options  opts) {
+int parse_options_delete(Parse_Options  opts)
+{
 	resources_delete(opts->resources);
 	xfree(opts, sizeof(struct Parse_Options_s));
 	return 0;
@@ -97,7 +99,7 @@ void parse_options_set_cost_model_type(Parse_Options opts, int cm)
 		opts->cost_model.compare_fn = &VDAL_compare_parse;
 		break;
 	default:
-		error("Illegal cost model: %d\n", cm);
+		prt_error("Error: Illegal cost model: %d\n", cm);
 	}
 }
 
@@ -256,11 +258,11 @@ int parse_options_get_display_postscript(Parse_Options opts) {
 void parse_options_set_display_constituents(Parse_Options opts, int dummy)
 {
 	if ((dummy < 0) || (dummy > 3)) {
-	   fprintf(stderr, "Possible values for constituents: \n");
-	   fprintf(stderr, "   0 (no display)\n");
-	   fprintf(stderr, "   1 (treebank style, multi-line indented)\n");
-	   fprintf(stderr, "   2 (flat tree, square brackets)\n");
-	   fprintf(stderr, "   3 (flat treebank style)\n");
+	   prt_error("Possible values for constituents: \n"
+	             "   0 (no display)\n"
+	             "   1 (treebank style, multi-line indented)\n"
+	             "   2 (flat tree, square brackets)\n"
+	             "   3 (flat treebank style)\n");
 	   opts->display_constituents = 0;
 	}
 	else opts->display_constituents = dummy;
@@ -338,15 +340,13 @@ dictionary_create(char * dict_name, char * pp_name,
                   char * cons_name, char * affix_name)
 {
 	Dictionary dict;
-	static int rand_table_inited=FALSE;
 	Dict_node *dict_node;
 
-	dict = (Dictionary) xalloc(sizeof(struct Dictionary_s));
+	init_memusage();
+	init_randtable();
 
-	if (!rand_table_inited) {
-		init_randtable();
-		rand_table_inited=TRUE;
-	}
+	dict = (Dictionary) xalloc(sizeof(struct Dictionary_s));
+	memset(dict, 0, sizeof(struct Dictionary_s));
 
 	dict->string_set = string_set_create();
 	dict->name = string_set_add(dict_name, dict->string_set);
@@ -358,22 +358,24 @@ dictionary_create(char * dict_name, char * pp_name,
 	dict->word_file_header = NULL;
 	dict->exp_list = NULL;
 	dict->affix_table = NULL;
+	dict->recursive_error = FALSE;
 
 	dict->fp = dictopen(dict->name, "r");
 	if (dict->fp == NULL) {
-		lperror(NODICT, dict_name);
+		prt_error("Error: Could not open dictionary %s\n", dict_name);
 		goto failure;
 	}
 
 	if (!read_dictionary(dict)) {
+		fclose(dict->fp);
 		goto failure;
 	}
+	fclose(dict->fp);
 
 	dict->affix_table = NULL;
 	if (affix_name != NULL) {
 		dict->affix_table = dictionary_create(affix_name, NULL, NULL, NULL);
 		if (dict->affix_table == NULL) {
-			fprintf(stderr, "%s\n", lperrmsg);
 			goto failure;
 		}
 	}
@@ -465,7 +467,7 @@ Dictionary dictionary_create_default_lang(void)
 int dictionary_delete(Dictionary dict)
 {
 	if (verbosity > 0) {
-		fprintf(stderr, "Freeing dictionary %s\n", dict->name);
+		prt_error("Info: Freeing dictionary %s\n", dict->name);
 	}
 
 	if (dict->affix_table != NULL) {
@@ -502,6 +504,7 @@ Sentence sentence_create(char *input_string, Dictionary dict)
 	int i;
 
 	sent = (Sentence) xalloc(sizeof(struct Sentence_s));
+	bzero(sent, sizeof(struct Sentence_s));
 	sent->dict = dict;
 	sent->length = 0;
 	sent->num_linkages_found = 0;
@@ -516,24 +519,29 @@ Sentence sentence_create(char *input_string, Dictionary dict)
 	sent->parse_info = NULL;
 	sent->string_set = string_set_create();
 
-	if (!separate_sentence(input_string, sent)) {
+	if (!separate_sentence(input_string, sent))
+	{
 		string_set_delete(sent->string_set);
 		xfree(sent, sizeof(struct Sentence_s));
 		return NULL;
 	}
 
+	error_report_set_sentence(sent);
+
 	sent->q_pruned_rules = FALSE; /* for post processing */
 	sent->is_conjunction = (char *) xalloc(sizeof(char)*sent->length);
 	set_is_conjunction(sent);
 	initialize_conjunction_tables(sent);
 
-	for (i=0; i<sent->length; i++) {
+	for (i=0; i<sent->length; i++)
+	{
 		/* in case we free these before they set to anything else */
 		sent->word[i].x = NULL;
 		sent->word[i].d = NULL;
 	}
 
-	if (!(dict->unknown_word_defined && dict->use_unknown_word)) {
+	if (!(dict->unknown_word_defined && dict->use_unknown_word))
+	{
 		if (!sentence_in_dictionary(sent)) {
 			sentence_delete(sent);
 			return NULL;
@@ -588,8 +596,11 @@ void sentence_delete(Sentence sent)
 	post_process_close_sentence(sent->dict->postprocessor);
 	free_deletable(sent);
 	free_effective_dist(sent);
+	free_count(sent);
+	free_analyze(sent);
 	xfree(sent->is_conjunction, sizeof(char)*sent->length);
 	xfree((char *) sent, sizeof(struct Sentence_s));
+	error_report_set_sentence(NULL);
 }
 
 int sentence_length(Sentence sent)
@@ -598,7 +609,8 @@ int sentence_length(Sentence sent)
 	return sent->length;
 }
 
-char * sentence_get_word(Sentence sent, int index) {
+char * sentence_get_word(Sentence sent, int index)
+{
 	if (!sent) return NULL;
 	return sent->word[index].string;
 }
@@ -680,10 +692,9 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 		/* We know that sent->num_linkages_found is bogus, possibly negative */
 		sent->num_linkages_found = opts->linkage_limit;
 		if (opts->verbosity > 1)
-		  fprintf(stdout,
-				  "Warning: Count overflow.\n"
-				  "Considering a random subset of %d of an unknown and large number of linkages\n",
-				  opts->linkage_limit);
+		  prt_error("Warning: Count overflow.\n"
+			  "Considering a random subset of %d of an unknown and large number of linkages\n",
+				opts->linkage_limit);
 	}
 	N_linkages_found = sent->num_linkages_found;
 
@@ -699,53 +710,63 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 	{
 		N_linkages_alloced = opts->linkage_limit;
 		if (opts->verbosity > 1) {
-		  fprintf(stdout,
-				  "Warning: Considering a random subset of %d of %d linkages\n",
+		  prt_error("Warning: Considering a random subset of %d of %d linkages\n",
 				  N_linkages_alloced, N_linkages_found);
 		}
 	}
 	else N_linkages_alloced = N_linkages_found;
 
-	link_info=(Linkage_info *)xalloc(N_linkages_alloced * sizeof(Linkage_info));
+	link_info = (Linkage_info *) xalloc(N_linkages_alloced * sizeof(Linkage_info));
 	N_linkages_post_processed = N_valid_linkages = 0;
 
-	/* generate an array of linkage indices to examine */
+	/* Generate an array of linkage indices to examine */
 	indices = (int *) xalloc(N_linkages_alloced * sizeof(int));
-	if (overflowed) {
-		for (in=0; in<N_linkages_alloced; in++) {
+	if (overflowed)
+	{
+		for (in=0; in < N_linkages_alloced; in++)
+		{
 			indices[in] = -(in+1);
 		}
 	}
-	else {
-		my_random_initialize(N_linkages_found + sent->length);
-		for (in=0; in<N_linkages_alloced; in++) {
+	else
+	{
+		sent->rand_state = N_linkages_found + sent->length;
+		for (in=0; in<N_linkages_alloced; in++)
+		{
 			denom = (double) N_linkages_alloced;
 			block_bottom = (int) (((double)in*(double) N_linkages_found)/denom);
 			block_top = (int) (((double)(in+1)*(double)N_linkages_found)/denom);
-			indices[in] = block_bottom + (my_random() % (block_top-block_bottom));
+			indices[in] = block_bottom + (rand_r(&sent->rand_state) % (block_top-block_bottom));
 		}
-		my_random_finalize();
 	}
 
 	only_canonical_allowed = (!(overflowed || (N_linkages_found > 2*opts->linkage_limit)));
-	/* When we're processing only a small subset of the linkages, don't worry
-	   about restricting the set we consider to be canonical ones.  In the extreme
-	   case where we are only generating 1 in a million linkages, it's very unlikely
-	   that we'll hit two symmetric variants of the same linkage anyway. */
-
+	/* When we're processing only a small subset of the linkages,
+	 * don't worry about restricting the set we consider to be
+	 * canonical ones.  In the extreme case where we are only 
+	 * generating 1 in a million linkages, it's very unlikely
+	 * that we'll hit two symmetric variants of the same linkage
+	 * anyway.
+	 */ 
 	/* (optional) first pass: just visit the linkages */
-	/* The purpose of these two passes is to make the post-processing more
-	   efficient.  Because (hopefully) by the time you do the real work
-	   in the 2nd pass you've pruned the relevant rule set in the first pass. */
-	if (sent->length >= opts->twopass_length) {
+	/* The purpose of these two passes is to make the post-processing
+	 * more efficient.  Because (hopefully) by the time you do the
+	 * real work in the 2nd pass you've pruned the relevant rule set
+	 * in the first pass.
+	 */
+	if (sent->length >= opts->twopass_length)
+	{
 		for (in=0; (in < N_linkages_alloced) &&
-				   (!resources_exhausted(opts->resources)); in++) {
+				   (!resources_exhausted(opts->resources)); in++)
+		{
 			extract_links(indices[in], sent->null_count, sent->parse_info);
-			if (set_has_fat_down(sent)) {
+			if (set_has_fat_down(sent))
+			{
 				if (only_canonical_allowed && !is_canonical_linkage(sent)) continue;
 				analyze_fat_linkage(sent, opts, PP_FIRST_PASS);
 			}
-			else {
+			else
+			{
 				analyze_thin_linkage(sent, opts, PP_FIRST_PASS);
 			}
 		}
@@ -753,9 +774,11 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 
 	/* second pass: actually perform post-processing */
 	for (in=0; (in < N_linkages_alloced) &&
-			   (!resources_exhausted(opts->resources)); in++) {
+			   (!resources_exhausted(opts->resources)); in++)
+	{
 		extract_links(indices[in], sent->null_count, sent->parse_info);
-		if (set_has_fat_down(sent)) {
+		if (set_has_fat_down(sent))
+		{
 			canonical = is_canonical_linkage(sent);
 			if (only_canonical_allowed && !canonical) continue;
 			link_info[N_linkages_post_processed] =
@@ -763,7 +786,8 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 			link_info[N_linkages_post_processed].fat = TRUE;
 			link_info[N_linkages_post_processed].canonical = canonical;
 		}
-		else {
+		else
+		{
 			link_info[N_linkages_post_processed] =
 				analyze_thin_linkage(sent, opts, PP_SECOND_PASS);
 			link_info[N_linkages_post_processed].fat = FALSE;
@@ -779,15 +803,30 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 	qsort((void *)link_info, N_linkages_post_processed, sizeof(Linkage_info),
 		  (int (*)(const void *, const void *)) opts->cost_model.compare_fn);
 
-	if (!resources_exhausted(opts->resources)) {
-		assert(! ((N_linkages_post_processed == 0) &&
-				  (N_linkages_found > 0) &&
-				  (N_linkages_found < opts->linkage_limit)),
-			   "None of the linkages is canonical");
+	if (!resources_exhausted(opts->resources))
+	{
+		if ((N_linkages_post_processed == 0) &&
+		    (N_linkages_found > 0) &&
+		    (N_linkages_found < opts->linkage_limit))
+		{
+			/* With the current parser, the following sentence will elicit
+			 * this error:
+			 *
+			 * Well, say, Joe, you can be Friar Tuck or Much the miller's
+			 * son, and lam me with a quarter-staff; or I'll be the Sheriff
+			 * of Nottingham and you be Robin Hood a little while and kill
+			 * me.
+			 */
+			prt_error("Error: None of the linkages is canonical\n"
+			          "\tN_linkages_post_processed=%d "
+			          "N_linkages_found=%d\n",
+			          N_linkages_post_processed,
+			          N_linkages_found);
+		}
 	}
 
 	if (opts->verbosity > 1) {
-		fprintf(stdout, "%d of %d linkages with no P.P. violations\n",
+		prt_error("Info: %d of %d linkages with no P.P. violations\n",
 				N_valid_linkages, N_linkages_post_processed);
 	}
 
@@ -817,6 +856,9 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 		return 0;
 	}
 
+	init_analyze(sent);
+	init_count(sent);
+
 	expression_prune(sent);
 	print_time(opts, "Finished expression pruning");
 	prepare_to_parse(sent, opts);
@@ -952,11 +994,13 @@ void linkage_delete(Linkage linkage)
 	exfree(linkage, sizeof(struct Linkage_s));
 }
 
-static int links_are_equal(Link l, Link m) {
+static int links_are_equal(Link *l, Link *m)
+{
 	return ((l->l == m->l) && (l->r == m->r) && (strcmp(l->name, m->name)==0));
 }
 
-static int link_already_appears(Linkage linkage, Link link, int a) {
+static int link_already_appears(Linkage linkage, Link *link, int a)
+{
 	int i, j;
 
 	for (i=0; i<a; ++i) {
@@ -985,7 +1029,7 @@ static Sublinkage unionize_linkage(Linkage linkage)
 {
 	int i, j, num_in_union=0;
 	Sublinkage u;
-	Link link;
+	Link *link;
 	const char *p;
 
 	for (i=0; i<linkage->num_sublinkages; ++i) {
@@ -995,7 +1039,7 @@ static Sublinkage unionize_linkage(Linkage linkage)
 		}
 	}
 
-	u.link = (Link *) exalloc(sizeof(Link)*num_in_union);
+	u.link = (Link **) exalloc(sizeof(Link *)*num_in_union);
 	u.num_links = num_in_union;
 	zero_sublinkage(&u);
 
@@ -1075,7 +1119,8 @@ int linkage_get_num_links(Linkage linkage) {
 	return linkage->sublinkage[current].num_links;
 }
 
-static int verify_link_index(Linkage linkage, int index) {
+static int verify_link_index(Linkage linkage, int index)
+{
 	int errno;
 	if ((index < 0) ||
 		(index >= linkage->sublinkage[linkage->current].num_links)) {
@@ -1085,8 +1130,9 @@ static int verify_link_index(Linkage linkage, int index) {
 	return 1;
 }
 
-int linkage_get_link_length(Linkage linkage, int index) {
-	Link link;
+int linkage_get_link_length(Linkage linkage, int index)
+{
+	Link *link;
 	int word_has_link[MAX_SENTENCE];
 	int i, length;
 	int current = linkage->current;
@@ -1111,22 +1157,25 @@ int linkage_get_link_length(Linkage linkage, int index) {
 	return length;
 }
 
-int linkage_get_link_lword(Linkage linkage, int index) {
-	Link link;
+int linkage_get_link_lword(Linkage linkage, int index)
+{
+	Link *link;
 	if (!verify_link_index(linkage, index)) return -1;
 	link = linkage->sublinkage[linkage->current].link[index];
 	return link->l;
 }
 
-int linkage_get_link_rword(Linkage linkage, int index) {
-	Link link;
+int linkage_get_link_rword(Linkage linkage, int index)
+{
+	Link *link;
 	if (!verify_link_index(linkage, index)) return -1;
 	link = linkage->sublinkage[linkage->current].link[index];
 	return link->r;
 }
 
-const char * linkage_get_link_label(Linkage linkage, int index) {
-	Link link;
+const char * linkage_get_link_label(Linkage linkage, int index)
+{
+	Link *link;
 	if (!verify_link_index(linkage, index)) return NULL;
 	link = linkage->sublinkage[linkage->current].link[index];
 	return link->name;
@@ -1134,7 +1183,7 @@ const char * linkage_get_link_label(Linkage linkage, int index) {
 
 const char * linkage_get_link_llabel(Linkage linkage, int index)
 {
-	Link link;
+	Link *link;
 	if (!verify_link_index(linkage, index)) return NULL;
 	link = linkage->sublinkage[linkage->current].link[index];
 	return link->lc->string;
@@ -1142,7 +1191,7 @@ const char * linkage_get_link_llabel(Linkage linkage, int index)
 
 const char * linkage_get_link_rlabel(Linkage linkage, int index)
 {
-	Link link;
+	Link *link;
 	if (!verify_link_index(linkage, index)) return NULL;
 	link = linkage->sublinkage[linkage->current].link[index];
 	return link->rc->string;
