@@ -16,7 +16,7 @@
 #include <link-grammar/api.h>
 #include "constituents.h"
 
-#define MAXCONSTITUENTS 1024
+#define MAXCONSTITUENTS 8192
 #define MAXSUBL 16
 #define OPEN_BRACKET '['
 #define CLOSE_BRACKET ']'
@@ -63,6 +63,7 @@ typedef struct
 	int templist[MAX_ELTS];
 	constituent_t constituent[MAXCONSTITUENTS];
 	andlist_t andlist[MAX_ANDS];
+	int r_limit;
 } con_context_t;
 
 /* ================================================================ */
@@ -112,9 +113,8 @@ static void adjust_for_right_comma(con_context_t *ctxt, Linkage linkage, int c)
 static void print_constituent(con_context_t *ctxt, Linkage linkage, int c)
 {
 	int w;
-	/* Sentence sent;
-	   sent = linkage_get_sentence(linkage); **PV* using linkage->word not sent->word */
-	if (verbosity<2) return;
+	if (verbosity < 2) return;
+
 	printf("  c %2d %4s [%c] (%2d-%2d): ",
 		   c, ctxt->constituent[c].type, ctxt->constituent[c].domain_type,
 		   ctxt->constituent[c].left, ctxt->constituent[c].right);
@@ -136,14 +136,16 @@ static void print_constituent(con_context_t *ctxt, Linkage linkage, int c)
  * type ctype2, call it c2. It then generates a new constituent of
  * ctype3, containing all the words in c2 but not c1.
  */
-static int gen_comp(con_context_t *ctxt, Linkage linkage, int numcon_total, int numcon_subl,
-					const char * ctype1, const char * ctype2, const char * ctype3, int x)
+static int gen_comp(con_context_t *ctxt, Linkage linkage,
+                    int numcon_total, int numcon_subl,
+					     const char * ctype1, const char * ctype2,
+                    const char * ctype3, int x)
 {
 	int w, w2, w3, c, c1, c2, done;
 	c = numcon_total + numcon_subl;
 
-	for (c1=numcon_total; c1<numcon_total + numcon_subl; c1++) {
-
+	for (c1=numcon_total; c1<numcon_total + numcon_subl; c1++)
+	{
 		/* If ctype1 is NP, it has to be an appositive to continue */
 		if ((x==4) && (post_process_match("MX#*", ctxt->constituent[c1].start_link)==0))
 			continue;
@@ -202,10 +204,10 @@ static int gen_comp(con_context_t *ctxt, Linkage linkage, int numcon_total, int
 		if (!(strcmp(ctxt->constituent[c1].type, ctype1)==0))
 			continue;
 
-		if (verbosity>=2)
+		if (verbosity >= 2)
 			printf("Generating complement constituent for c %d of type %s\n",
 				   c1, ctype1);
-		done=0;
+		done = 0;
 		for (w2=ctxt->constituent[c1].left; (done==0) && (w2>=0); w2--) {
 			for (w3=ctxt->constituent[c1].right; w3<linkage->num_words; w3++) {
 				for (c2=numcon_total; (done==0) &&
@@ -272,15 +274,20 @@ static int gen_comp(con_context_t *ctxt, Linkage linkage, int numcon_total, int
 						print_constituent(ctxt, linkage, c);
 					}
 					c++;
-					assert(c < MAXCONSTITUENTS, "Too many constituents");
+					if (MAXCONSTITUENTS <= c)
+					{
+						prt_error("Error: Too many constituents (a).\n");
+						c--;
+					}
 					done = 1;
 				}
 			}
 		}
-		if (verbosity>=2) {
-		  if (done==0)
-			printf("No constituent added, because no larger %s " \
-				   " was found\n", ctype2);
+		if (verbosity >= 2)
+		{
+			if (done == 0)
+				printf("No constituent added, because no larger %s " \
+					   " was found\n", ctype2);
 		}
 	}
 	numcon_subl = c - numcon_total;
@@ -405,8 +412,8 @@ static int find_next_element(con_context_t *ctxt,
 		}
 		if (ok == 0) continue;
 
-		ctxt->templist[num_elements]=c;
-		addedone=1;
+		ctxt->templist[num_elements] = c;
+		addedone = 1;
 		num_lists = find_next_element(ctxt, linkage, c, numcon_total,
 									  num_elements+1, num_lists);
 	}
@@ -417,7 +424,20 @@ static int find_next_element(con_context_t *ctxt,
 			ctxt->andlist[num_lists].num = num_elements;
 		}
 		num_lists++;
-		assert(num_lists < MAX_ANDS, "Constituent overflowed andlist!\n");
+		if (MAX_ANDS <= num_lists)
+		{
+			/* With the current parser, the following will cause an
+			 * overflow:
+			 *
+			 * I have not seen the grysbok, or the suni, or the dibitag, or
+			 * the lechwi, or the aoul, or the gerenuk, or the blaauwbok,
+			 * or the chevrotain, or lots of others, but who in the world
+			 * could guess what they were or what they looked like, judging
+			 * only from the names?
+			 */
+			prt_error("Error: Constituent overflowed andlist!\n");
+			return (num_lists-1); 
+		}
 	}
 	return num_lists;
 }
@@ -436,8 +456,9 @@ static int merge_constituents(con_context_t *ctxt, Linkage linkage, int numcon_t
 		/* Find and invalidate any constituents with negative length */
 		if(ctxt->constituent[c1].right < ctxt->constituent[c1].left)
 		{
-		  if(verbosity >= 2) printf("WARNING: Constituent %d has negative length. Deleting it.\n", c1);
-		  ctxt->constituent[c1].valid = 0;
+			if(verbosity >= 2)
+				prt_error("Warning: Constituent %d has negative length. Deleting it.\n", c1);
+			ctxt->constituent[c1].valid = 0;
 		}
 		ctxt->constituent[c1].canon = c1;
 	}
@@ -597,7 +618,7 @@ static int merge_constituents(con_context_t *ctxt, Linkage linkage, int numcon_t
 					if (!((ctxt->constituent[c2].left<=ctxt->constituent[c1].left) &&
 						  (ctxt->constituent[c2].right>=ctxt->constituent[c1].right)))
 						continue;
-					if (verbosity>=2)
+					if (verbosity >= 2)
 						printf("Found that c%d in list %d is bigger " \
 							   "than c%d in list %d\n", c2, n2, c1, n);
 					ok=1;
@@ -607,7 +628,8 @@ static int merge_constituents(con_context_t *ctxt, Linkage linkage, int numcon_t
 					   contains ALL the elements of n.
 					   If not, n is invalid. */
 
-					for (a3=0; a3 < ctxt->andlist[n].num; a3++) {
+					for (a3=0; a3 < ctxt->andlist[n].num; a3++)
+					{
 						c3 = ctxt->andlist[n].e[a3];
 						if ((ctxt->constituent[c2].left>ctxt->constituent[c3].left) ||
 							(ctxt->constituent[c2].right<ctxt->constituent[c3].right))
@@ -616,7 +638,8 @@ static int merge_constituents(con_context_t *ctxt, Linkage linkage, int numcon_t
 					if (ok != 0)
 						continue;
 					ctxt->andlist[n].valid = 0;
-					if (verbosity>=2) {
+					if (verbosity >= 2)
+					{
 						printf("Eliminating andlist, " \
 							   "n=%d, a=%d, n2=%d, a2=%d: ",
 							   n, a, n2, a2);
@@ -631,7 +654,8 @@ static int merge_constituents(con_context_t *ctxt, Linkage linkage, int numcon_t
 	}
 
 
-	if (verbosity>=2) {
+	if (verbosity >= 2)
+	{
 		printf("And-lists after pruning:\n");
 		for (n=0; n<num_lists; n++) {
 			if (ctxt->andlist[n].valid==0)
@@ -878,20 +902,24 @@ static int last_minute_fixes(con_context_t *ctxt, Linkage linkage, int numcon_to
 	/* Check once more to see if constituents are nested (checking BETWEEN sublinkages
 	   this time) */
 
-	while (1) {
+	while (1)
+	{
 		adjustment_made=0;
-		for (c=0; c<numcon_total; c++) {
+		for (c=0; c<numcon_total; c++)
+		{
 			if(ctxt->constituent[c].valid==0) continue;
-			for (c2=0; c2<numcon_total; c2++) {
+			for (c2=0; c2<numcon_total; c2++)
+			{
 				if(ctxt->constituent[c2].valid==0) continue;
 				if ((ctxt->constituent[c].left < ctxt->constituent[c2].left) &&
 					(ctxt->constituent[c].right < ctxt->constituent[c2].right) &&
-					(ctxt->constituent[c].right >= ctxt->constituent[c2].left)) {
-
-					if (verbosity>=2) {
-					  printf("WARNING: the constituents aren't nested! Adjusting them." \
-							   "(%d, %d)\n", c, c2);
-					  }
+					(ctxt->constituent[c].right >= ctxt->constituent[c2].left))
+				{
+					if (verbosity >= 2)
+					{
+						prt_error("Warning: the constituents aren't nested! "
+						          "Adjusting them. (%d, %d)\n", c, c2);
+					}
 					ctxt->constituent[c].left = ctxt->constituent[c2].left;
 				}
 			}
@@ -913,18 +941,23 @@ static void count_words_used(con_context_t *ctxt, Linkage linkage)
 	num_subl = linkage->num_sublinkages;
 	if(linkage->unionized==1 && num_subl>1) num_subl--;
 
-	if (verbosity>=2)
+	if (verbosity >= 2)
 		printf("Number of sublinkages = %d\n", num_subl);
-	for (i=0; i<num_subl; i++) {
+
+	for (i=0; i<num_subl; i++)
+	{
 		for (w=0; w<linkage->num_words; w++) ctxt->word_used[i][w]=0;
 		linkage->current=i;
-		for (link=0; link<linkage_get_num_links(linkage); link++) {
+		for (link=0; link<linkage_get_num_links(linkage); link++)
+		{
 			ctxt->word_used[i][linkage_get_link_lword(linkage, link)]=1;
 			ctxt->word_used[i][linkage_get_link_rword(linkage, link)]=1;
 		}
-		if (verbosity>=2) {
+		if (verbosity >= 2)
+		{
 			printf("Sublinkage %d: ", i);
-			for (w=0; w<linkage->num_words; w++) {
+			for (w=0; w<linkage->num_words; w++)
+			{
 				if (ctxt->word_used[i][w]==0) printf("0 ");
 				if (ctxt->word_used[i][w]==1) printf("1 ");
 			}
@@ -933,8 +966,6 @@ static void count_words_used(con_context_t *ctxt, Linkage linkage)
 	}
 }
 
-static int r_limit=0;
-
 static int add_constituent(con_context_t *ctxt, int c, Linkage linkage, Domain domain,
                            int l, int r, const char * name)
 {
@@ -942,7 +973,7 @@ static int add_constituent(con_context_t *ctxt, int c, Linkage linkage, Domain d
 
 	/* Avoid running off end, to walls. */
 	if (l < 1) l=1;
-	if (r > r_limit) r = r_limit;
+	if (r > ctxt->r_limit) r = ctxt->r_limit;
 	assert(l <= r, "negative constituent length!" );
 
 	ctxt->constituent[c].left = l;
@@ -997,8 +1028,8 @@ static const char * cons_of_domain(char domain_type)
 	case 'z':
 		return "VP";
 	default:
-		printf("Illegal domain: %c\n", domain_type);
-		assert(0, "Illegal domain");
+		prt_error("Error: Illegal domain: %c\n", domain_type);
+		return "";
 	}
 }
 
@@ -1012,7 +1043,7 @@ static int read_constituents_from_domains(con_context_t *ctxt, Linkage linkage,
 	const char * name;
 	Domain domain;
 
-	r_limit = linkage->num_words-2; /**PV**/
+	ctxt->r_limit = linkage->num_words-2; /**PV**/
 
 	subl = &linkage->sublinkage[s];
 
@@ -1121,7 +1152,7 @@ static int read_constituents_from_domains(con_context_t *ctxt, Linkage linkage,
 			if (ctxt->wordtype[w] == QDTYPE) {
 				/* Now find the finite verb to the right, start an S */
 				/* Limit w2 to sentence length. */
-				// for( w2=w+1; w2 < r_limit-1; w2++ )
+				// for( w2=w+1; w2 < ctxt->r_limit-1; w2++ )
 				for (w2 = w+1; w2 < rightmost; w2++)
 				  if ((ctxt->wordtype[w2] == STYPE) || (ctxt->wordtype[w2] == PTYPE)) break;
 
@@ -1132,10 +1163,10 @@ static int read_constituents_from_domains(con_context_t *ctxt, Linkage linkage,
 		}
 
 		if (ctxt->constituent[c].domain_type=='\0') {
-			error("Error: no domain type assigned to constituent\n");
+			prt_error("Error: no domain type assigned to constituent\n");
 		}
 		if (ctxt->constituent[c].start_link==NULL) {
-			error("Error: no type assigned to constituent\n");
+			prt_error("Error: no type assigned to constituent\n");
 		}
 	}
 
@@ -1217,21 +1248,22 @@ static int read_constituents_from_domains(con_context_t *ctxt, Linkage linkage,
 					if ((strcmp(linkage->word[ctxt->constituent[c2].right], ",")==0) ||
 						(strcmp(linkage->word[ctxt->constituent[c2].right],
 								"RIGHT-WALL")==0)) {
-						if (verbosity>=2)
+						if (verbosity >= 2)
 							printf("Adjusting %d to fix comma overlap\n", c2);
 						adjust_for_right_comma(ctxt, linkage, c2);
 						adjustment_made=1;
 					}
 					else if (strcmp(linkage->word[ctxt->constituent[c].left], ",")==0) {
-						if (verbosity>=2)
+						if (verbosity >= 2)
 							printf("Adjusting c %d to fix comma overlap\n", c);
 						adjust_for_left_comma(ctxt, linkage, c);
 						adjustment_made=1;
 					}
 					else {
-					  if (verbosity>=2) {
-						printf("WARNING: the constituents aren't nested! Adjusting them." \
-							   "(%d, %d)\n", c, c2);
+						if (verbosity >= 2)
+						{
+							prt_error("Warning: the constituents aren't nested! "
+							          "Adjusting them. (%d, %d)\n", c, c2);
 					  }
 					  ctxt->constituent[c].left = ctxt->constituent[c2].left;
 					}
@@ -1273,59 +1305,90 @@ static int read_constituents_from_domains(con_context_t *ctxt, Linkage linkage,
 
 static char * exprint_constituent_structure(con_context_t *ctxt, Linkage linkage, int numcon_total)
 {
+	int have_opened = 1;
 	int c, w;
 	int leftdone[MAXCONSTITUENTS];
 	int rightdone[MAXCONSTITUENTS];
 	int best, bestright, bestleft;
 	Sentence sent;
 	char s[100], * p;
-	String * cs = String_create();
+	String * cs = string_new();
 
-	assert (numcon_total < MAXCONSTITUENTS, "Too many constituents");
+	assert (numcon_total < MAXCONSTITUENTS, "Too many constituents (b)");
 	sent = linkage_get_sentence(linkage);
 
-	for(c=0; c<numcon_total; c++) {
+	for (c = 0; c < numcon_total; c++)
+	{
 		leftdone[c]=0;
 		rightdone[c]=0;
 	}
 
-	if(verbosity>=2)
+	if (verbosity >= 2)
 		printf("\n");			
 
-	for(w=1; w<linkage->num_words; w++) {	
+	for (w=1; w<linkage->num_words; w++)
+	{
 		/* Skip left wall; don't skip right wall, since it may
 		   have constituent boundaries */
 
-		while(1) {
+		while(1)
+		{
 			best = -1;
 			bestright = -1;
-			for(c=0; c<numcon_total; c++) {
-				if ((ctxt->constituent[c].left==w) &&
-					(leftdone[c]==0) && (ctxt->constituent[c].valid==1) &&
+			for (c=0; c < numcon_total; c++)
+			{
+				if ((ctxt->constituent[c].left == w) &&
+					(leftdone[c] == 0) && (ctxt->constituent[c].valid == 1) &&
 					(ctxt->constituent[c].right >= bestright)) {
 					best = c;
 					bestright = ctxt->constituent[c].right;
 				}
 			}
-			if (best==-1)
+			if (best == -1)
 				break;
-			leftdone[best]=1;
-			if(ctxt->constituent[best].aux==1) continue;
+
+			leftdone[best] = 1;
+			/* have_open is a hack to avoid printing anything until 
+			 * bracket is opened */
+			if (w == 1) have_opened = 0; 
+			if (ctxt->constituent[best].aux == 1) continue;
+			have_opened = 1;
 			append_string(cs, "%c%s ", OPEN_BRACKET, ctxt->constituent[best].type);
 		}
 
-		if (w<linkage->num_words-1) {
-			/* Don't print out right wall */
+		/* Don't print out right wall */
+		if (have_opened && (w < linkage->num_words - 1))
+		{
+			char *p;
 			strcpy(s, sent->word[w].string);
 
+			/* Constituent processing will crash if the sentence contains
+			 * square brackets, so we have to do something ... replace 
+			 * them with curly braces ... will have to do.
+			 */
+			p = strchr(s, OPEN_BRACKET);
+			while(p)
+			{
+				*p = '{';
+				p = strchr(p, OPEN_BRACKET);
+			}
+
+			p = strchr(s, CLOSE_BRACKET);
+			while(p)
+			{
+				*p = '}';
+				p = strchr(p, CLOSE_BRACKET);
+			}
+
 			/* Now, if the first character of the word was
 			   originally uppercase, we put it back that way */
-			if (sent->word[w].firstupper ==1 )
+			if (sent->word[w].firstupper == 1)
 				upcase_utf8_str(s, s, MAX_WORD);
 			append_string(cs, "%s ", s);
 		}
 
-		while(1) {
+		while(1)
+		{
 			best = -1;
 			bestleft = -1;
 			for(c=0; c<numcon_total; c++) {
@@ -1346,14 +1409,12 @@ static char * exprint_constituent_structure(con_context_t *ctxt, Linkage linkage
 	}
 
 	append_string(cs, "\n");
-	p = exalloc(strlen(cs->p)+1);
-	strcpy(p, cs->p);
-	exfree(cs->p, sizeof(char)*cs->allocated);
-	exfree(cs, sizeof(String));
+	p = string_copy(cs);
+	string_delete(cs);
 	return p;
 }
 
-static char * print_flat_constituents(con_context_t *ctxt, Linkage linkage)
+static char * do_print_flat_constituents(con_context_t *ctxt, Linkage linkage)
 {
 	int num_words;
 	Sentence sent;
@@ -1369,27 +1430,52 @@ static char * print_flat_constituents(con_context_t *ctxt, Linkage linkage)
 	count_words_used(ctxt, linkage);
 
 	num_subl = linkage->num_sublinkages;
-	if(num_subl > MAXSUBL) {
-	  num_subl=MAXSUBL;
-	  if(verbosity>=2) printf("Number of sublinkages exceeds maximum: only considering first %d sublinkages\n", MAXSUBL);
+	if (num_subl > MAXSUBL)
+	{
+	  num_subl = MAXSUBL;
+	  if (verbosity >= 2)
+		printf("Number of sublinkages exceeds maximum: only considering first %d sublinkages\n", MAXSUBL);
 	}
-	if(linkage->unionized==1 && num_subl>1) num_subl--;
-	for (s=0; s<num_subl; s++) {
+	if (linkage->unionized==1 && num_subl>1) num_subl--;
+	for (s=0; s<num_subl; s++)
+	{
 		linkage_set_current_sublinkage(linkage, s);
 		linkage_post_process(linkage, pp);
 		num_words = linkage_get_num_words(linkage);
 		generate_misc_word_info(ctxt, linkage);
 		numcon_subl = read_constituents_from_domains(ctxt, linkage, numcon_total, s);
 		numcon_total = numcon_total + numcon_subl;
+		if (MAXCONSTITUENTS <= numcon_total)
+		{
+			prt_error("Error: Too many constituents (c).\n");
+			numcon_total = MAXCONSTITUENTS-1;
+			break;
+		}
 	}
 	numcon_total = merge_constituents(ctxt, linkage, numcon_total);
+	if (MAXCONSTITUENTS <= numcon_total)
+	{
+		prt_error("Error: Too many constituents (d).\n");
+		numcon_total = MAXCONSTITUENTS-1;
+	}
 	numcon_total = last_minute_fixes(ctxt, linkage, numcon_total);
+	if (MAXCONSTITUENTS <= numcon_total)
+	{
+		prt_error("Error: Too many constituents (e).\n");
+		numcon_total = MAXCONSTITUENTS-1;
+	}
 	q = exprint_constituent_structure(ctxt, linkage, numcon_total);
 	string_set_delete(ctxt->phrase_ss);
 	ctxt->phrase_ss = NULL;
 	return q;
 }
 
+static char * print_flat_constituents(Linkage linkage)
+{
+	con_context_t ctxt;
+	return do_print_flat_constituents(&ctxt, linkage);
+}
+
 static CType token_type (char *token)
 {
 	if ((token[0] == OPEN_BRACKET) && (strlen(token)>1))
@@ -1399,7 +1485,8 @@ static CType token_type (char *token)
 	return WORD_TOK;
 }
 
-static CNode * make_CNode(char *q) {
+static CNode * make_CNode(char *q)
+{
 	CNode * cn;
 	cn = exalloc(sizeof(CNode));
 	cn->label = (char *) exalloc(sizeof(char)*(strlen(q)+1));
@@ -1450,33 +1537,59 @@ static void print_tree(String * cs, int indent, CNode * n, int o1, int o2)
 	int i, child_offset;
 	CNode * m;
 
-	if (n==NULL) return;
+	if (n == NULL) return;
 
 	if (indent)
-		for (i=0; i<o1; ++i)
+		for (i = 0; i < o1; ++i)
 			append_string(cs, " ");
 	append_string(cs, "(%s ", n->label);
-	child_offset = o2 + strlen(n->label)+2;
+	child_offset = o2 + strlen(n->label) + 2;
+
+	for (m = n->child; m != NULL; m = m->next)
+	{
+		if (m->child == NULL)
+		{
+			char * p;
+			/* If the original string has left or right parens in it,
+			 * the printed string will be messed up by these ... 
+			 * so replace them by curly braces. What else can one do?
+			 */
+			p = strchr(m->label, '(');
+			while(p)
+			{
+				*p = '{';
+				p = strchr(p, '(');
+			}
+
+			p = strchr(m->label, ')');
+			while(p)
+			{
+				*p = '}';
+				p = strchr(p, ')');
+			}
 
-	for (m=n->child; m!=NULL; m=m->next) {
-		if (m->child == NULL) {
 			append_string(cs, "%s", m->label);
 			if ((m->next != NULL) && (m->next->child == NULL))
 				append_string(cs, " ");
 		}
-		else {
-			if (m != n->child) {
+		else
+		{
+			if (m != n->child)
+			{
 				if (indent) append_string(cs, "\n");
 				else append_string(cs, " ");
 				print_tree(cs, indent, m, child_offset, child_offset);
 			}
-			else {
+			else
+			{
 				print_tree(cs, indent, m, 0, child_offset);
 			}
-			if ((m->next != NULL) && (m->next->child == NULL)) {
-				if (indent) {
+			if ((m->next != NULL) && (m->next->child == NULL))
+			{
+				if (indent)
+				{
 					append_string(cs, "\n");
-					for (i=0; i<child_offset; ++i)
+					for (i = 0; i < child_offset; ++i)
 						append_string(cs, " ");
 				}
 				else append_string(cs, " ");
@@ -1509,11 +1622,8 @@ CNode * linkage_constituent_tree(Linkage linkage)
 	char *p, *q, *saveptr;
 	int len;
 	CNode * root;
-	con_context_t *ctxt;
 
-	ctxt = (con_context_t *) malloc(sizeof(con_context_t));
-	p = print_flat_constituents(ctxt, linkage);
-	free(ctxt);
+	p = print_flat_constituents(linkage);
 
 	len = strlen(p);
 	q = strtok_r(p, " ", &saveptr);
@@ -1554,27 +1664,18 @@ char * linkage_print_constituent_tree(Linkage linkage, int mode)
 	}
 	else if (mode == 1 || mode == 3)
 	{
-		cs = String_create();
+		cs = string_new();
 		root = linkage_constituent_tree(linkage);
 		print_tree(cs, (mode==1), root, 0, 0);
 		linkage_free_constituent_tree(root);
 		append_string(cs, "\n");
-		p = exalloc(strlen(cs->p)+1);
-		strcpy(p, cs->p);
-		exfree(cs->p, sizeof(char)*cs->allocated);
-		exfree(cs, sizeof(String));
+		p = string_copy(cs);
+		string_delete(cs);
 		return p;
 	}
 	else if (mode == 2)
 	{
-		char * str;
-		con_context_t *ctxt;
-
-		ctxt = (con_context_t *) malloc(sizeof(con_context_t));
-		str = print_flat_constituents(ctxt, linkage);
-		free(ctxt);
-
-		return str;
+		return print_flat_constituents(linkage);
 	}
 	assert(0, "Illegal mode in linkage_print_constituent_tree");
 	return NULL;
