@@ -15,17 +15,16 @@
 #include <stdarg.h>
 #include <link-grammar/api.h>
 
-static int center[MAX_SENTENCE];
-static int N_words_to_print;  /* version of N_words in this file for printing links */
 const char * trailer(int mode);
 const char * header(int mode);
 
-static void set_centers(Linkage linkage, int print_word_0)
+static void set_centers(Linkage linkage, int center[], int print_word_0, int N_words_to_print)
 {
 	int i, len, tot;
 	tot = 0;
 	if (print_word_0) i=0; else i=1;
-	for (; i<N_words_to_print; i++) {
+	for (; i < N_words_to_print; i++)
+	{
 		/* Centers obtained by counting the characters, 
 		 * not the bytes in the string.
 		 * len = strlen(linkage->word[i]);
@@ -37,12 +36,14 @@ static void set_centers(Linkage linkage, int print_word_0)
 }
 
 /* the following are all for generating postscript */
-static int link_heights[MAX_LINKS];
-/* tells the height of the links above the sentence */
-static int row_starts[MAX_SENTENCE];
-/* the word beginning each row of the display */
-static int N_rows;
-/* the number of rows */
+typedef struct
+{
+	int N_rows; /* N_rows -- the number of rows */
+	/* tells the height of the links above the sentence */
+	int link_heights[MAX_LINKS];
+	/* the word beginning each row of the display */
+	int row_starts[MAX_SENTENCE];
+} ps_ctxt_t;
 
 /**
  * prints s then prints the last |t|-|s| characters of t.
@@ -99,7 +100,7 @@ char * linkage_print_links_and_domains(Linkage linkage)
 {
 	int link, longest, j;
 	int N_links = linkage_get_num_links(linkage);
-	String * s = String_create();
+	String * s = string_new();
 	char * links_string;
 	char ** dname;
 
@@ -127,27 +128,28 @@ char * linkage_print_links_and_domains(Linkage linkage)
 		append_string(s, "        %s\n\n", linkage_get_violation_name(linkage));
 	}
 	
-	links_string = exalloc(strlen(s->p)+1);
-	strcpy(links_string, s->p);
-	exfree(s->p, sizeof(char)*s->allocated);
-	exfree(s, sizeof(String));
-
+	links_string = string_copy(s);
+	string_delete(s);
 	return links_string; 
 }
 
-static char * build_linkage_postscript_string(Linkage linkage) {
+/**
+ */
+static char * build_linkage_postscript_string(Linkage linkage, ps_ctxt_t *pctx)
+{
 	int link, i,j;
 	int d;
 	int print_word_0 = 0, print_word_N = 0, N_wall_connectors, suppressor_used;
-	Sublinkage *sublinkage=&(linkage->sublinkage[linkage->current]);
+	Sublinkage *sublinkage = &(linkage->sublinkage[linkage->current]);
 	int N_links = sublinkage->num_links;
-	Link *ppla = sublinkage->link;
+	Link **ppla = sublinkage->link;
 	String  * string;
 	char * ps_string;
 	Dictionary dict = linkage->sent->dict;
 	Parse_Options opts = linkage->opts;
+	int N_words_to_print;
 
-	string = String_create();
+	string = string_new();
 
 	N_wall_connectors = 0;
 	if (dict->left_wall_defined) {
@@ -186,7 +188,6 @@ static char * build_linkage_postscript_string(Linkage linkage) {
 		print_word_N = TRUE;
 	}
 
-
 	if (print_word_0) d=0; else d=1;
 
 	i = 0;
@@ -212,7 +213,7 @@ static char * build_linkage_postscript_string(Linkage linkage) {
 		j++;
 		append_string(string,"[%d %d %d",
 				ppla[link]->l-d, ppla[link]->r-d, 
-				link_heights[link]);
+				pctx->link_heights[link]);
 		if (ppla[link]->lc->label < 0) {
 			append_string(string," (%s)]", ppla[link]->name);
 		} else {
@@ -222,17 +223,15 @@ static char * build_linkage_postscript_string(Linkage linkage) {
 	append_string(string,"]");
 	append_string(string,"\n");
 	append_string(string,"[");
-	for (j=0; j<N_rows; j++ ){
-		if (j>0) append_string(string, " %d", row_starts[j]);
-		else append_string(string,"%d", row_starts[j]);
+	for (j=0; j < pctx->N_rows; j++ )
+	{
+		if (j>0) append_string(string, " %d", pctx->row_starts[j]);
+		else append_string(string,"%d", pctx->row_starts[j]);
 	}
 	append_string(string,"]\n");
 
-	ps_string = exalloc(strlen(string->p)+1);
-	strcpy(ps_string, string->p);
-	exfree(string->p, sizeof(char)*string->allocated);
-	exfree(string, sizeof(String));
-
+	ps_string = string_copy(string);
+	string_delete(string);
 	return ps_string; 
 }
 
@@ -249,7 +248,8 @@ void compute_chosen_words(Sentence sent, Linkage linkage)
 	const char * chosen_words[MAX_SENTENCE];
 	Parse_Options opts = linkage->opts;
 
-	for (i=0; i<sent->length; i++) {   /* get rid of those ugly ".Ixx" */
+	for (i=0; i<sent->length; i++)
+	{   /* get rid of those ugly ".Ixx" */
 		chosen_words[i] = sent->word[i].string;
 		if (pi->chosen_disjuncts[i] == NULL) {  
 			/* no disjunct is used on this word because of null-links */
@@ -293,31 +293,34 @@ void compute_chosen_words(Sentence sent, Linkage linkage)
 
 #define MAX_HEIGHT 30
 
-static char picture[MAX_HEIGHT][MAX_LINE];
-static char xpicture[MAX_HEIGHT][MAX_LINE];
-
 /** 
  * String allocated with exalloc.  
  * Needs to be freed with linkage_free_diagram()
  */
-char * linkage_print_diagram(Linkage linkage)
+static char * linkage_print_diagram_ctxt(Linkage linkage, ps_ctxt_t *pctx)
 {
 	int i, j, k, cl, cr, row, top_row, width, flag;
 	const char *s;
 	char *t;
 	int print_word_0 = 0, print_word_N = 0, N_wall_connectors, suppressor_used;
+	int center[MAX_SENTENCE];
 	char connector[MAX_TOKEN_LENGTH];
 	int line_len, link_length;
 	Sublinkage *sublinkage=&(linkage->sublinkage[linkage->current]);
 	int N_links = sublinkage->num_links;
-	Link *ppla = sublinkage->link;
+	Link **ppla = sublinkage->link;
 	String * string;
 	char * gr_string;
 	Dictionary dict = linkage->sent->dict;
 	Parse_Options opts = linkage->opts;
 	int x_screen_width = parse_options_get_screen_width(opts);
+	int N_words_to_print;
 
-	string = String_create();
+	char picture[MAX_HEIGHT][MAX_LINE];
+	char xpicture[MAX_HEIGHT][MAX_LINE];
+
+
+	string = string_new();
 
 	N_wall_connectors = 0;
 	if (dict->left_wall_defined) {
@@ -360,7 +363,7 @@ char * linkage_print_diagram(Linkage linkage)
 	N_words_to_print = linkage->num_words;
 	if (!print_word_N) N_words_to_print--;
 	
-	set_centers(linkage, print_word_0);
+	set_centers(linkage, center, print_word_0, N_words_to_print);
 	line_len = center[N_words_to_print-1]+1;
 	
 	for (k=0; k<MAX_HEIGHT; k++) {
@@ -390,14 +393,12 @@ char * linkage_print_diagram(Linkage linkage)
 			}
 			/* we know it fits, so put it in this row */
 
-			link_heights[j] = row;
+			pctx->link_heights[j] = row;
 			
 			if (2*row+2 > MAX_HEIGHT-1) {
 				append_string(string, "The diagram is too high.\n");
-				gr_string = exalloc(strlen(string->p)+1);
-				strcpy(gr_string, string->p);
-				exfree(string->p, sizeof(char)*string->allocated);
-				exfree(string, sizeof(String));
+				gr_string = string_copy(string);
+				string_delete(string);
 				return gr_string; 
 			}
 			if (row > top_row) top_row = row;
@@ -497,9 +498,9 @@ char * linkage_print_diagram(Linkage linkage)
 	
 	if (print_word_0) i = 0; else i = 1;
 	k = 0;
-	N_rows = 0;
-	row_starts[N_rows] = 0;
-	N_rows++;
+	pctx->N_rows = 0;
+	pctx->row_starts[pctx->N_rows] = 0;
+	pctx->N_rows++;
 	while(i < N_words_to_print) {
 		append_string(string, "\n");
 		width = 0;
@@ -508,8 +509,8 @@ char * linkage_print_diagram(Linkage linkage)
 			i++;
 		} while((i<N_words_to_print) &&
 			  (width + ((int)strlen(linkage->word[i]))+1 < x_screen_width));
-		row_starts[N_rows] = i - (!print_word_0);    /* PS junk */
-		if (i<N_words_to_print) N_rows++;     /* same */
+		pctx->row_starts[pctx->N_rows] = i - (!print_word_0);    /* PS junk */
+		if (i<N_words_to_print) pctx->N_rows++;     /* same */
 		for (row = top_row; row >= 0; row--) {
 			flag = TRUE;
 			for (j=k;flag&&(j<k+width)&&(xpicture[row][j]!='\0'); j++){
@@ -525,24 +526,38 @@ char * linkage_print_diagram(Linkage linkage)
 		append_string(string, "\n");
 		k += width;
 	}
-	gr_string = exalloc(strlen(string->p)+1);
-	strcpy(gr_string, string->p);
-	exfree(string->p, sizeof(char)*string->allocated);
-	exfree(string, sizeof(String));
+	gr_string = string_copy(string);
+	string_delete(string);
 	return gr_string; 
 }
 
+char * linkage_print_diagram(Linkage linkage)
+{
+	ps_ctxt_t ctx;
+	return linkage_print_diagram_ctxt(linkage, &ctx);
+}
+
 void linkage_free_diagram(char * s)
 {
 		exfree(s, strlen(s)+1);
 }
 
+void linkage_free_links_and_domains(char * s)
+{
+		exfree(s, strlen(s)+1);
+}
+
 char * linkage_print_postscript(Linkage linkage, int mode)
 {
 	char * ps, * qs;
 	int size;
 
-	ps = build_linkage_postscript_string(linkage);
+	/* call the ascii printer to initialize the row size stuff. */
+	ps_ctxt_t ctx;
+	char * ascii = linkage_print_diagram_ctxt(linkage, &ctx);
+	linkage_free_diagram(ascii);
+
+	ps = build_linkage_postscript_string(linkage, &ctx);
 	size = strlen(header(mode)) + strlen(ps) + strlen(trailer(mode)) + 1;
 	
 	qs = exalloc(sizeof(char)*size);
