@@ -90,26 +90,48 @@
 static int link_advance(Dictionary dict);
 static Dict_node * abridged_lookup_list(Dictionary dict, const char *s);
 
-static void dict_error(Dictionary dict, const char * s)
+static void dict_error2(Dictionary dict, const char * s, const char *s2)
 {
 	int i;
 	char tokens[1024], t[128];
 
+	if (dict->recursive_error) return;
+	dict->recursive_error = TRUE;
+
 	tokens[0] = '\0';
 	for (i=0; i<5 && dict->token[0] != '\0' ; i++)
 	{
 		sprintf(t, "\"%s\" ", dict->token);
 		strcat(tokens, t);
-		(void) link_advance(dict);
+		link_advance(dict);
 	}
-	lperror(DICTPARSE, ". %s\n\t line %d, tokens = %s\n",
-			s, dict->line_number, tokens);
+	if (s2)
+	{
+		prt_error("Error parsing dictionary %s.\n"
+		          "%s %s\n\t line %d, tokens = %s\n",
+		        dict->name,
+		        s, s2, dict->line_number, tokens);
+	}
+	else
+	{
+		prt_error("Error parsing dictionary %s.\n"
+		          "%s\n\t line %d, tokens = %s\n",
+		        dict->name,
+		        s, dict->line_number, tokens);
+	}
+	dict->recursive_error = FALSE;
+}
+
+static void dict_error(Dictionary dict, const char * s)
+{
+	dict_error2(dict, s, NULL);
 }
 
 static void warning(Dictionary dict, const char * s)
 {
-	printf("\nWarning: %s\n",s);
-	printf("line %d, current token = \"%s\"\n", dict->line_number, dict->token);
+	prt_error("Warning: %s\n"
+	          "\tline %d, current token = \"%s\"\n",
+	          s, dict->line_number, dict->token);
 }
 
 Exp * Exp_create(Dictionary dict)
@@ -152,7 +174,7 @@ static wint_t get_character(Dictionary dict, int quote_mode)
 static int link_advance(Dictionary dict)
 {
 	wint_t c;
-	int i;
+	int nr, i;
 	int quote_mode;
 
 	dict->is_special = FALSE;
@@ -189,8 +211,19 @@ static int link_advance(Dictionary dict)
 				dict_error(dict, "White space inside of token");
 				return 0;
 			}
+
 			/* store UTF8 internally, always. */
-			i += wctomb(&dict->token[i], c);
+			nr = wctomb(&dict->token[i], c);
+			if (nr < 0) {
+#ifndef _WIN32
+				dict_error2(dict, "Unable to read UTF8 string in current locale",
+				         nl_langinfo(CODESET));
+#else
+				dict_error(dict, "Unable to read UTF8 string in current locale");
+#endif
+				return 0;
+			}
+			i += nr;
 		} else {
 			if (strchr(SPECIAL, c) != NULL) {
 				if (i == 0) {
@@ -220,7 +253,17 @@ static int link_advance(Dictionary dict)
 				quote_mode = TRUE;
 			} else {
 				/* store UTF8 internally, always. */
-				i += wctomb(&dict->token[i], c);
+				nr = wctomb_check(&dict->token[i], c);
+				if (nr < 0) {
+#ifndef _WIN32
+					dict_error2(dict, "Unable to read UTF8 string in current locale",
+					         nl_langinfo(CODESET));
+#else
+					dict_error(dict, "Unable to read UTF8 string in current locale");
+#endif
+					return 0;
+				}
+				i += nr;
 			}
 		}
 		c = get_character(dict, quote_mode);
@@ -238,7 +281,8 @@ static int is_equal(Dictionary dict, wint_t c)
 	        dict->token[1] == '\0');
 }
 
-static int check_connector(Dictionary dict, const char * s) {
+static int check_connector(Dictionary dict, const char * s)
+{
 	/* makes sure the string s is a valid connector */
 	int i;
 	i = strlen(s);
@@ -327,7 +371,8 @@ static Exp * connector(Dictionary dict)
  * This creates a node with one child (namely e).  Initializes
  * the cost to zero.
  */
-Exp * make_unary_node(Dictionary dict, Exp * e) {
+Exp * make_unary_node(Dictionary dict, Exp * e)
+{
 	Exp * n;
 	n = Exp_create(dict);
 	n->type = AND_type;  /* these must be AND types */
@@ -342,7 +387,8 @@ Exp * make_unary_node(Dictionary dict, Exp * e) {
  * This creates a node with zero children.  Initializes
  * the cost to zero.
  */
-static Exp * make_zeroary_node(Dictionary dict) {
+static Exp * make_zeroary_node(Dictionary dict)
+{
 	Exp * n;
 	n = Exp_create(dict);
 	n->type = AND_type;  /* these must be AND types */
@@ -351,11 +397,13 @@ static Exp * make_zeroary_node(Dictionary dict) {
 	return n;
 }
 
-static Exp * make_optional_node(Dictionary dict, Exp * e) {
-/* This creates an OR node with two children, one the given node,
-   and the other as zeroary node.  This has the effect of creating
-   what used to be called an optional node.
-*/
+/**
+ * This creates an OR node with two children, one the given node,
+ * and the other as zeroary node.  This has the effect of creating
+ * what used to be called an optional node.
+ */
+static Exp * make_optional_node(Dictionary dict, Exp * e)
+{
 	Exp * n;
 	E_list *el, *elx;
 	n = Exp_create(dict);
@@ -373,10 +421,13 @@ Exp * expression(Dictionary dict);
 
 #if ! defined INFIX_NOTATION
 
-Exp * operator_exp(Dictionary dict, int type) {
-/* We're looking at the first of the stuff after an "and" or "or".	 */
-/* Build a Exp node for this expression.  Set the cost and optional	*/
-/* fields to the default values.  Set the type field according to type */
+/** 
+ * We're looking at the first of the stuff after an "and" or "or".
+ * Build a Exp node for this expression.  Set the cost and optional
+ * fields to the default values.  Set the type field according to type
+ */
+Exp * operator_exp(Dictionary dict, int type)
+{
 	Exp * n;
 	E_list first;
 	E_list * elist;
@@ -405,7 +456,8 @@ Exp * operator_exp(Dictionary dict, int type) {
  * Looks for the stuff that is allowed to be inside of parentheses
  * either & or | followed by a list, or a terminal symbol.
  */
-Exp * in_parens(Dictionary dict) {
+Exp * in_parens(Dictionary dict)
+{
 	Exp * e;
 
 	if (is_equal(dict, '&') || (strcmp(token, "and")==0)) {
@@ -423,11 +475,13 @@ Exp * in_parens(Dictionary dict) {
 	}
 }
 
-Exp * expression(Dictionary dict) {
-/* Build (and return the root of) the tree for the expression beginning	 */
-/* with the current token.  At the end, the token is the first one not part */
-/* of this expression.													  */
-
+/** 
+ * Build (and return the root of) the tree for the expression beginning
+ * with the current token.  At the end, the token is the first one not
+ * part of this expression.
+ */
+Exp * expression(Dictionary dict)
+{
 	Exp * n;
 	if (is_equal(dict, '(')) {
 		if (!link_advance(dict)) {
@@ -488,14 +542,18 @@ Exp * expression(Dictionary dict) {
 
 Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok);
 
-Exp * expression(Dictionary dict) {
-/* Build (and return the root of) the tree for the expression beginning  */
-/* with the current token.  At the end, the token is the first one not   */
-/* part of this expression.											  */
+/**
+ * Build (and return the root of) the tree for the expression beginning
+ * with the current token.  At the end, the token is the first one not
+ * part of this expression.
+ */
+Exp * expression(Dictionary dict)
+{
 	return restricted_expression(dict, TRUE,TRUE);
 }
 
-Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok) {
+Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok)
+{
 	Exp * nl=NULL, * nr, * n;
 	E_list *ell, *elr;
 	if (is_equal(dict, '(')) {
@@ -617,7 +675,8 @@ Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok) {
 
 /* verbose version */
 /*
-int dict_compare(char *s, char *t) {
+int dict_compare(char *s, char *t)
+{
 	int ss, tt;
 	while (*s != '\0' && *s == *t) {
 		s++;
@@ -702,21 +761,22 @@ static void insert_list(Dictionary dict, Dict_node * p, int l)
 	} 
 	else if (is_idiom_word(dn->string))
 	{
-		printf("*** Word \"%s\" found near line %d.\n", dn->string, dict->line_number);
-		printf("	Words ending \".Ix\" (x a number) are reserved for idioms.\n");
-		printf("	This word will be ignored.\n");
+		prt_error("Warning: Word \"%s\" found near line %d.\n"
+		        "\tWords ending \".Ix\" (x a number) are reserved for idioms.\n"
+		        "\tThis word will be ignored.\n",
+		        dn->string, dict->line_number);
 		xfree((char *)dn, sizeof(Dict_node));
 	}
 	else if ((dn_head = abridged_lookup_list(dict, dn->string))!= NULL)
 	{
 		Dict_node *dnx;
-		printf("*** The word \"%s\"", dn->string);
-		printf(" found near line %d of %s matches the following words:\n",
-			   dict->line_number, dict->name);
+		prt_error("Warning: The word \"%s\" "
+		          "found near line %d of %s matches the following words:\n",
+	             dn->string, dict->line_number, dict->name);
 		for (dnx = dn_head; dnx != NULL; dnx = dnx->right) {
-			printf(" %s", dnx->string);
+			fprintf(stderr, "\t%s", dnx->string);
 		}
-		printf("\n	This word will be ignored.\n");
+		fprintf(stderr, "\n\tThis word will be ignored.\n");
 		free_lookup_list(dn_head);
 		xfree((char *)dn, sizeof(Dict_node));
 	}
@@ -731,55 +791,74 @@ static void insert_list(Dictionary dict, Dict_node * p, int l)
 }
 
 /**
+ * read_entry() -- read one dictionary entry
  * Starting with the current token parse one dictionary entry.
  * Add these words to the dictionary.
  */
 static int read_entry(Dictionary dict)
 {
-    Exp *n;
-    int i;
+	Exp *n;
+	int i;
 
-	Dict_node  *dn_new, *dnx, *dn = NULL;
+	Dict_node *dn_new, *dnx, *dn = NULL;
 
-	for (; !is_equal(dict, ':') ; link_advance(dict)) {
-		  if (dict->is_special) {
-			  dict_error(dict, "I expected a word but didn\'t get it.");
-			  return 0;
-		  }
-		  if (dict->token[0] == '/') {
-			  /* it's a word-file name */
-			  dn = read_word_file(dict, dn, dict->token);
-			  if (dn == NULL) {
-				  lperror(WORDFILE, " %s\n", dict->token);
-				  return 0;
-			  }
-		  } else {
-			  dn_new = (Dict_node *) xalloc(sizeof(Dict_node));
-			  dn_new->left = dn;
-			  dn = dn_new;
-			  dn->file = NULL;
-			  dn->string = string_set_add(dict->token, dict->string_set);
-		  }
+	while (!is_equal(dict, ':'))
+	{
+		if (dict->is_special)
+		{
+			dict_error(dict, "I expected a word but didn\'t get it.");
+			return 0;
+		}
+
+		/* if it's a word-file name */
+		if (dict->token[0] == '/') {
+			dn = read_word_file(dict, dn, dict->token);
+			if (dn == NULL)
+			{
+				prt_error("Error opening word file %s\n", dict->token);
+				return 0;
+			}
+		} else {
+			dn_new = (Dict_node *) xalloc(sizeof(Dict_node));
+			dn_new->left = dn;
+			dn = dn_new;
+			dn->file = NULL;
+			dn->string = string_set_add(dict->token, dict->string_set);
+		}
+
+		/* Advance to next entry, unless error */
+		if (0 == link_advance(dict)) return 0;
 	}
-	if (!link_advance(dict)) {  /* pass the : */
+
+	/* pass the : */
+	if (!link_advance(dict))
+	{
 		return 0;
 	}
+
 	n = expression(dict);
-	if (n == NULL) {
+	if (n == NULL)
+	{
 		return 0;
 	}
-	if (!is_equal(dict, ';')) {
+
+	if (!is_equal(dict, ';'))
+	{
 		dict_error(dict, "Expecting \";\" at the end of an entry.");
 		return 0;
 	}
-	if (!link_advance(dict)) {  /* pass the ; */
+
+	/* pass the ; */
+	if (!link_advance(dict))
+	{
 		return 0;
 	}
 
-	/* at this point, dn points to a list of Dict_nodes connected by	  */
-	/* their left pointers.  These are to be inserted into the dictionary */
+	/* at this point, dn points to a list of Dict_nodes connected by */
+	/* their left pointers. These are to be inserted into the dictionary */
 	i = 0;
-	for (dnx = dn; dnx != NULL; dnx = dnx->left) {
+	for (dnx = dn; dnx != NULL; dnx = dnx->left)
+	{
 		dnx->exp = n;
 		i++;
 	}
@@ -787,7 +866,8 @@ static int read_entry(Dictionary dict)
 	return 1;
 }
 
-void print_expression(Exp * n) {
+void print_expression(Exp * n)
+{
 	E_list * el; int i;
 	if (n == NULL) {
 		printf("NULL expression");
@@ -810,7 +890,8 @@ void print_expression(Exp * n) {
 	}
 }
 
-static void rprint_dictionary_data(Dict_node * n) {
+static void rprint_dictionary_data(Dict_node * n)
+{
 		if (n==NULL) return;
 		rprint_dictionary_data(n->left);
 		printf("%s: ", n->string);
@@ -819,25 +900,24 @@ static void rprint_dictionary_data(Dict_node * n) {
 		rprint_dictionary_data(n->right);
 }
 
-void print_dictionary_data(Dictionary dict) {
+void print_dictionary_data(Dictionary dict)
+{
 	rprint_dictionary_data(dict->root);
 }
 
-
 int read_dictionary(Dictionary dict)
 {
-	lperrno = 0;
-	if (!link_advance(dict)) {
-		fclose(dict->fp);
+	if (!link_advance(dict))
+	{
 		return 0;
 	}
-	while(dict->token[0] != '\0') {
-			if (!read_entry(dict)) {
-			fclose(dict->fp);
+	while (dict->token[0] != '\0')
+	{
+		if (!read_entry(dict))
+		{
 			return 0;
 		}
 	}
-	fclose(dict->fp);
 	return 1;
 }
 
@@ -1023,39 +1103,40 @@ int boolean_abridged_lookup(Dictionary dict, const char *s)
 }
 
 /* the following functions are for handling deletion */
-
-static Dict_node * parent;
-static Dict_node * to_be_deleted;
-
-
-static int find_one_non_idiom_node(Dict_node * p, Dict_node * dn, const char * s) {
-/* Returns true if it finds a non-idiom dict_node in a file that matches
-   the string s.
-
-   ** note: this now DOES include non-file words in its search.
-
-	Also sets parent and to_be_deleted appropriately.
-*/
+/**
+ * Returns true if it finds a non-idiom dict_node in a file that matches
+ * the string s.
+ *
+ ** note: this now DOES include non-file words in its search.
+ *
+ * Also sets parent and to_be_deleted appropriately.
+ */
+static int find_one_non_idiom_node(Dict_node * p, Dict_node * dn,
+                                   const char * s,
+                                   Dict_node **parent, Dict_node **to_be_deleted)
+{
 	int m;
 	if (dn == NULL) return FALSE;
 	m = dict_match(s, dn->string);
 	if (m <= 0) {
-		if (find_one_non_idiom_node(dn,dn->left, s)) return TRUE;
+		if (find_one_non_idiom_node(dn,dn->left, s, parent, to_be_deleted)) return TRUE;
 	}
 /*	if ((m == 0) && (!is_idiom_word(dn->string)) && (dn->file != NULL)) { */
 	if ((m == 0) && (!is_idiom_word(dn->string))) {
-		to_be_deleted = dn;
-		parent = p;
+		*to_be_deleted = dn;
+		*parent = p;
 		return TRUE;
 	}
 	if (m >= 0) {
-		if (find_one_non_idiom_node(dn,dn->right, s)) return TRUE;
+		if (find_one_non_idiom_node(dn,dn->right, s, parent, to_be_deleted)) return TRUE;
 	}
 	return FALSE;
 }
 
 static void set_parent_of_node(Dictionary dict,
-						Dict_node *p, Dict_node * del, Dict_node * newnode)
+                               Dict_node *p,
+                               Dict_node * del,
+                               Dict_node * newnode)
 {
 	if (p == NULL) {
 		dict->root = newnode;
@@ -1070,12 +1151,16 @@ static void set_parent_of_node(Dictionary dict,
 	}
 }
 
-int delete_dictionary_words(Dictionary dict, const char * s) {
-/* This deletes all the non-idiom words of the dictionary that match
-   the given string.  Returns TRUE if some deleted, FALSE otherwise.
-*/
+/**
+ * This deletes all the non-idiom words of the dictionary that match
+ * the given string.  Returns TRUE if some deleted, FALSE otherwise.
+ */
+int delete_dictionary_words(Dictionary dict, const char * s)
+{
 	Dict_node *pred, *pred_parent;
-	if (!find_one_non_idiom_node(NULL, dict->root, s)) return FALSE;
+	Dict_node *parent, *to_be_deleted;
+
+	if (!find_one_non_idiom_node(NULL, dict->root, s, &parent, &to_be_deleted)) return FALSE;
 	for(;;) {
 		/* now parent and to_be_deleted are set */
 		if (to_be_deleted->file != NULL) {
@@ -1097,11 +1182,12 @@ int delete_dictionary_words(Dictionary dict, const char * s) {
 			set_parent_of_node(dict, pred_parent, pred, pred->left);
 			xfree((char *)pred, sizeof(Dict_node));
 		}
-		if (!find_one_non_idiom_node(NULL, dict->root, s)) return TRUE;
+		if (!find_one_non_idiom_node(NULL, dict->root, s, &parent, &to_be_deleted)) return TRUE;
 	}
 }
 
-static void free_Word_file(Word_file * wf) {
+static void free_Word_file(Word_file * wf)
+{
 	Word_file *wf1;
 
 	for (;wf != NULL; wf = wf1) {
@@ -1110,11 +1196,13 @@ static void free_Word_file(Word_file * wf) {
 	}
 }
 
-/* The following two functions free the Exp s and the
-   E_lists of the dictionary.  Not to be confused with
-   free_E_list in utilities.c
+/**
+ * The following two functions free the Exp s and the
+ * E_lists of the dictionary.  Not to be confused with
+ * free_E_list in utilities.c
  */
-static void free_Elist(E_list * l) {
+static void free_Elist(E_list * l)
+{
 	E_list * l1;
 
 	for (; l != NULL; l = l1) {
@@ -1123,7 +1211,8 @@ static void free_Elist(E_list * l) {
 	}
 }
 
-static void free_Exp_list(Exp * e) {
+static void free_Exp_list(Exp * e)
+{
 	Exp * e1;
 	for (; e != NULL; e = e1) {
 		e1 = e->next;
