@@ -13,10 +13,19 @@
 #ifndef _LINK_GRAMMAR_UTILITIES_H_
 #define _LINK_GRAMMAR_UTILITIES_H_
 
+#ifndef _WIN32
+#include <langinfo.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
+
+#ifndef __CYGWIN__
 #include <wchar.h>
 #include <wctype.h>
+#endif
+
+#include "error.h"
 
 #ifndef FALSE
 #define FALSE 0
@@ -26,24 +35,26 @@
 #define TRUE 1
 #endif
 
-#ifdef CYGWIN
+#ifdef __CYGWIN__
 #define _WIN32 1
-#endif /* CYGWIN */
+#endif /* __CYGWIN__ */
 
 #ifdef _WIN32
 #include <windows.h>
 
+#ifndef _MSC_VER
+   /* CYGWIN/MINGW compilers don't support the "inline" keyword. */
+#define inline
+#endif
+
 #ifdef _MSC_VER
 #define snprintf _snprintf
 #endif
 
-/* Windows compilers don't support the "inline" keyword. */
-#define inline
-
 /* CYGWIN on Windows doesn't have UTF8 support, or wide chars ... 
  * However, MS Visual C appearnetly does ... 
  */
-#ifdef CYGWIN
+#ifdef __CYGWIN__
 #define mbtowc(w,s,n)  ({*((char *)(w)) = *(s); 1;})
 #define wctomb(s,w)    ({*((char *)(s)) = ((char)(w)); 1;})
 #define iswupper  isupper
@@ -54,6 +65,8 @@
 #define wint_t    int
 #define fgetwc    fgetc
 #define WEOF      EOF
+#define towlower  tolower
+#define towupper  toupper
 #endif
 
 /* strtok_r is missing in windows */
@@ -61,6 +74,20 @@ char * strtok_r (char *s, const char *delim, char **saveptr);
 
 #endif /* _WIN32 */
 
+static inline int wctomb_check(char *s, wchar_t wc)
+{
+	int nr = wctomb(s, wc);
+	if (nr < 0) {
+#ifndef _WIN32
+		prt_error("Fatal Error: unknwon character set %s\n", nl_langinfo(CODESET));
+#else
+		prt_error("Fatal Error: unknown character set\n");
+#endif
+		exit(1);
+	}
+	return nr;
+}
+
 static inline int is_utf8_upper(const char *s)
 {
 	wchar_t c;
@@ -134,19 +161,19 @@ void safe_strcpy(char *u, const char * v, size_t usize);
 void safe_strcat(char *u, const char *v, size_t usize);
 char *safe_strdup(const char *u);
 
-void xfree(void *, int);
-void exfree(void *, int);
 void init_randtable(void);
-int  next_power_of_two_up(int);
 void left_print_string(FILE* fp, const char *, const char *);
 
-void my_random_initialize(int seed);
-void my_random_finalize(void);
-int  my_random(void);
+/* routines for allocating basic objects */
+void init_memusage(void);
+void * xalloc(size_t);
+void * xrealloc(void *, size_t oldsize, size_t newsize);
+void xfree(void *, size_t);
+void * exalloc(size_t);
+void exfree(void *, size_t);
+size_t get_space_in_use(void);
+size_t get_max_space_used(void);
 
-/* routines for copying basic objects */
-void *      xalloc(int);
-void *      exalloc(int);
 
 char * get_default_locale(void);
 char * join_path(const char * prefix, const char * suffix);
@@ -154,4 +181,14 @@ char * join_path(const char * prefix, const char * suffix);
 FILE *dictopen(const char *filename, const char *how);
 void set_data_dir(const char * path);
 
+/**
+ * Returns the smallest power of two that is at least i and at least 1
+ */
+static inline int next_power_of_two_up(int i)
+{
+   int j=1;
+   while(j<i) j = j<<1;
+   return j;
+}
+
 #endif
