@@ -28,8 +28,8 @@ int  size_of_expression(Exp *);
 /* routines for copying basic objects */
 Disjunct *  copy_disjunct(Disjunct * );
 void        exfree_connectors(Connector *);
-Link        excopy_link(Link);
-void        exfree_link(Link);
+Link *      excopy_link(Link *);
+void        exfree_link(Link *);
 Connector * copy_connectors(Connector *);
 Connector * excopy_connectors(Connector * c);
 Disjunct *  catenate_disjuncts(Disjunct *, Disjunct *);
@@ -46,12 +46,11 @@ void set_is_conjunction(Sentence sent);
 int sentence_contains_conjunction(Sentence sent);
 int conj_in_range(Sentence sent, int lw, int rw);
 int max_link_length(Connector * c);
-void string_delete(char * p);
 
 /* Connector_set routines */
 Connector_set * connector_set_create(Exp *e);
 void connector_set_delete(Connector_set * conset);
-int match_in_connector_set(Connector_set *conset, Connector * c, int d);
+int match_in_connector_set(Sentence, Connector_set *conset, Connector * c, int d);
 
 Dict_node * list_whole_dictionary(Dict_node *, Dict_node *);
 int word_has_connector(Dict_node *, const char *, int);
